#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import json

import logging

from collections import defaultdict

from sqlalchemy import or_, and_, func
from sqlalchemy.orm import aliased, joinedload
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import IntegrityError
from zope.sqlalchemy import mark_changed

from ..db.model import Attr,\
        User,\
        Node,\
        Link,\
        ResourceGroup,\
        Network,\
        Project,\
        Scenario,\
        TemplateType,\
        ResourceAttr,\
        TypeAttr,\
        ResourceAttrMap,\
        ResourceScenario,\
        Dataset,\
        AttrGroup,\
        AttrGroupItem, \
        Dimension, \
        Unit


from .. import db
from .. import config

from ..exceptions import HydraError, ResourceNotFoundError
from ..util.permissions import required_perms, required_role

from . import units
from .objects import JSONObject
from .cache import cache

log = logging.getLogger(__name__)


def _get_network(network_id):
    try:
        network_i = db.DBSession.query(Network).filter(Network.id == network_id).one()
    except NoResultFound:
        raise HydraError("Network %s not found" % (network_id))
    return network_i

def _get_project(project_id):
    try:
        project_i = db.DBSession.query(Project).filter(Project.id == project_id).one()
    except NoResultFound:
        raise HydraError("Project %s not found" % (project_id))
    return project_i


def _get_resource(ref_key, ref_id):
    try:
        if ref_key == 'NODE':
            return db.DBSession.query(Node).filter(Node.id == ref_id).one()
        elif ref_key == 'LINK':
            return db.DBSession.query(Link).filter(Link.id == ref_id).one()
        elif ref_key == 'GROUP':
            return db.DBSession.query(ResourceGroup).filter(ResourceGroup.id == ref_id).one()
        elif ref_key == 'NETWORK':
            return _get_network(ref_id)
        elif ref_key == 'SCENARIO':
            return db.DBSession.query(Scenario).filter(Scenario.id == ref_id).one()
        elif ref_key == 'PROJECT':
            return _get_project(ref_id)
        return None
    except NoResultFound:
        raise ResourceNotFoundError("Resource %s with ID %s not found"%(ref_key, ref_id))

def _get_resource_id(ra):
    if ra.resource_id is not None:
        return ra.resource_id

    ref_key = ra.ref_key
    if ref_key == 'NETWORK':
        return ra.network_id
    elif ref_key == 'NODE':
        return ra.node_id
    elif ref_key == 'LINK':
        return ra.link_id
    elif ref_key == 'GROUP':
        return ra.group_id
    elif ref_key == 'PROJECT':
        return ra.project_id

def get_attribute_by_id(attr_id, **kwargs):
    """
        Get a specific attribute by its ID.
    """
    try:
        attr_i = db.DBSession.query(Attr).filter(Attr.id == attr_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Attribute (attribute id=%s) does not exist"%(attr_id))

    return attr_i

def get_attributes_by_id(attr_ids, **kwargs):
    """
        Get a list of specific attributes by their IDs.
    """
    if not attr_ids:
        return []

    return db.DBSession.query(Attr).filter(Attr.id.in_(attr_ids)).all()

def get_template_attributes(template_id, **kwargs):
    """
        Get a specific attribute by its ID.
    """

    try:
        attrs_i = db.DBSession.query(Attr).filter(
            TemplateType.template_id == template_id).filter(
                TypeAttr.type_id == TemplateType.id).filter(
                    Attr.id == TypeAttr.id).all()

        log.debug(attrs_i)
        return attrs_i
    except NoResultFound:
        return None


def get_attribute_by_name_and_dimension(name, dimension_id=None, network_id=None, project_id=None, **kwargs):
    """
        Get all attributes with the specified name and dimension, irrespective of
        scoping.
        dimension_id can be None, because in attribute the dimension_id is not anymore mandatory
        args:
            name (str): The name of the attribute. Lower() is called on this for comparison, so this
                        is case-insensitive
            dimension_id (int): the ID of the dimension of the attribute
        returns:
            list: JSONObjects derived from the Sqlalchemy rows.
    """

    log.info("Retrieving all attributes with name %s and dimension %s (network_id=%s) (project_id=%s)", name, dimension_id, network_id, project_id)
    try:
        attr_qry = db.DBSession.query(Attr).filter(
            and_(
                func.lower(Attr.name) == name.strip().lower(),
                Attr.dimension_id == dimension_id,
                Attr.network_id == network_id,
                Attr.project_id == project_id
            )
        )

        attr_i = attr_qry.first()

        log.debug("Attribute retrieved")
        return attr_i
    except NoResultFound:
        return None

def search_attributes(name, network_id=None, project_id=None, **kwargs):
    """
        Search for all attributes matching the given name, ignoring case
    """
    user_id = kwargs.get('user_id')
    name = name.lower()
    try:
        proj_attrs_i = []
        if project_id is not None:
            proj_i = db.DBSession.query(Project).filter(Project.id==project_id).one()
            proj_i.check_read_permission(user_id)

            proj_attrs_i = proj_i.get_scoped_attributes(name_match=name, include_hierarchy=True)
        attrs_dict = {a.name:a for a in proj_attrs_i}

        #Then get network-scoped attributes in case there are any scoped to the proejct
        #which supercede the project attributes
        net_attrs_i = []
        if network_id is not None:
            net_i = db.DBSession.query(Network).filter(Network.id==network_id).one()
            net_i.check_read_permission(user_id)
            include_network_hierarchy=True
            if project_id is not None:
                include_network_hierarchy=False
            net_attrs_i = net_i.get_scoped_attributes(name_match=name, include_hierarchy=include_network_hierarchy)

        for na in net_attrs_i:
            attrs_dict[na.name] = na

        #Finally add in all the global attributes which do not have the same
        #name as the scoped attributes. WHy? Becuase we assume that within scoping,
        #names must be unique --- you can't have a 'cost' at different dimensions within a scope,
        #so if there is a 'cost' which is scoped, then all 'cost' (regardless of dimension) can be ignored.
        global_attrs_i = db.DBSession.query(Attr).filter(
            func.lower(Attr.name).like(f'%{name}%'),
            Attr.network_id==None,
            Attr.project_id==None).all()

        for a in global_attrs_i:
            if a.name not in attrs_dict:
                attrs_dict[a.name] = JSONObject(a)

        return_attrs = []
        #now load the dimension for each attribute.
        for a_j in attrs_dict.values():
            if a_j.dimension_id is not None:
                dimension_i = db.DBSession.query(Dimension).filter(Dimension.id==a_j.dimension_id).one()
                a_j.dimension = JSONObject(dimension_i)
            return_attrs.append(a_j)
        log.debug("%s attributes matching %s", len(return_attrs), name)
        return return_attrs
    except NoResultFound:
        return None


def _add_attribute(attr, user_id, flush=True, do_reassign=False):
    """
    Add an attribute to the DB
    args:
        attr: A JSONObject representing the attr
        user_id: The ID of the user adding the attribute
        flush: Flag to indicate whether this should call the DB flush
        do_reassign: Flag to indicate whether any attributes scoped lower than the
                     incoming attribute should be removed. **WARNING*** this is
                     just here for testing purposes
     returns:
        JSONObject of new attr
    """
    log.debug("Adding attribute: %s", attr.name)

    attr_i = Attr(
        name=attr.name,
        dimension_id=attr.dimension_id,
        description=attr.description,
        network_id = attr.network_id,
        project_id = attr.project_id
    )

    _check_can_add_attribute(attr.name, attr.dimension_id, attr.project_id, attr.network_id)

    if attr.network_id is not None and attr.project_id is not None:
        raise HydraError(f"Unable to add attrubute {attr.name}. "+
                         "An attribute cannot have both a project_id and network_id")

    #users can only add attributes to networks or projects which they own.
    if attr.network_id is not None:
        net = _get_network(attr.network_id)
        if net.check_write_permission(user_id):
            attr_i.network_id = attr.network_id

    if attr.project_id is not None:
        proj = _get_project(attr.project_id)
        proj.check_write_permission(user_id)
        attr_i.project_id = attr.project_id

    #Only admins can add global attributes.
    if attr.network_id is None and attr.project_id is None:
        user = db.DBSession.query(User).filter(User.id == user_id).one()
        if not user.is_admin():
            raise PermissionError(f"User {user.username} does not have permission to add a global attribute."+
                                  "Please specify a network_id or project_id to the attribute.")

    db.DBSession.add(attr_i)
    if flush is True:
        db.DBSession.flush()

    log.info("New attr added")

    """
      Now that we have an ID, check for inconsistencies in the attribute scoping
      hierarchy, and fix them by removing any conflicting entries and reassigning
      any resource attributes to the non-conflicting attribute.
      Only do this for attributes not scoped to a network as a network is the lowest scope.
    """
    if do_reassign is True and attr_i.network_id is None:
        _reassign_scoped_attributes(attr_i.id, user_id)
        if flush is True:
            db.DBSession.flush()

    return JSONObject(attr_i)

def _reassign_scoped_attributes(attr_id, user_id):
    """
    If a matching attribute exists (same name & dimension) but scoped at a lower
    level, then we need to delete those attributes and then
    re-assign all resource attributes to use the new, higher-level attribute
    """
    attr_i = db.DBSession.query(Attr).filter(Attr.id == attr_id).one()

    """
      If a matching attribute exists (same name & dimension) but scoped at a lower
      level, then we need to delete those attributes, add the new attribute and then
      re-assign all resource attributes to use the new, global attribute
    """
    matching_attrs = get_attributes_by_name_and_dimension(
        attr_i.name,
        attr_i.dimension_id
    )
    if len(matching_attrs) == 1:
        #Only 1 returned value means this attr. More than 1 means there's a scoped
        #attribute with the same name and dimension
        assert matching_attrs[0].id == attr_id
        return

    #Reassign all resource attributes which point to scoped attirbutes, and then delete
    #the scoped attributes.
    scoped_resource_attrs_qry = db.DBSession.query(ResourceAttr).join(Attr).filter(
        ResourceAttr.attr_id == Attr.id,
        func.lower(Attr.name) == attr_i.name.lower(),
        Attr.dimension_id == attr_i.dimension_id,
        Attr.id != attr_id
    )


    """
      If this is a project scoped attribute then we only want to change the scope
      of attributes scoped to networks contained within this project, and leave
      other projects alone.
      If it's global, we want to stipulate that we want all attributes which
      are project-scoped also.
    """
    scoped_resource_attrs = []
    project_scope = []
    if attr_i.project_id is not None:
        """
            A `project_scope` is defined here which includes the ids of the current
            and any nested projects. This represents the scope within which matching
            network and project attrs will be rescoped to the current attr.
        """

        #first look up the hierarchy to see if there is an attribute scoped at a higher level.


        max_levels = int(config.get("limits", "project_max_nest_depth", 32))
        attr_proj = db.DBSession.query(Project).filter(Project.id == attr_i.project_id).one()
        child_projects = attr_proj.get_child_projects(user_id=user_id, levels=max_levels)
        project_scope = {p["id"] for p in child_projects} | {attr_i.project_id}
        network_scoped_resource_attrs_qry = scoped_resource_attrs_qry.join(Network).filter(
            Attr.network_id == Network.id,
            Network.project_id.in_(project_scope)
        )

        project_scoped_resource_attrs_qry = scoped_resource_attrs_qry.join(Project).filter(
            Attr.project_id.in_(project_scope)
        )

        scoped_resource_attrs = network_scoped_resource_attrs_qry.all() + project_scoped_resource_attrs_qry.all()

    if len(scoped_resource_attrs) > 0:
        log.info(f"{len(scoped_resource_attrs)} scoped attributes found with same name & dimension. Reassigning.")

    rescoped_from_networks = set()
    # Reassign the attributes, noting the network_id of of any network attrs
    for scoped_ra in scoped_resource_attrs:
        if scoped_ra.attr.network_id:
            rescoped_from_networks.add(scoped_ra.attr.network_id)
        scoped_ra.attr_id = attr_i.id

    # Remove rescoped attrs, both from project scope and those rescoped from network attrs
    for matching_attr in matching_attrs:
        if matching_attr.id != attr_id and (matching_attr.project_id in project_scope or matching_attr.network_id in rescoped_from_networks):
            db.DBSession.delete(matching_attr)

def _check_can_add_attribute(name, dimension, project_id, network_id, do_raise=True):
    """
        Check if an attribute can be added. If an attribute exists at a higher level
        (such as global) then it cannot be added to a lower scope.
        i.e. if a project-scoped attribute exists with a name of 'flow' and
        dimension of 'volumetric flow rate', then a network-scoped attribute with
        this name and dimenaion cannot be added, as the network scope is within the project scope.
    """

    if network_id is not None:
        net = _get_network(network_id)

        #look for an attribute with the same name and dimension but defined globally
        globally_scoped_attribute = get_attribute_by_name_and_dimension(name, dimension)
        if globally_scoped_attribute is not None:
            if do_raise is True:
                raise HydraError(
                    f"Unable to add attribute with name '{name}' and dimension '{dimension}' "
                    f"to network '{network_id}' as an "
                    f"attribute with this name and dimension already "
                    f"exists globally")
            else:
                return False

        #look for an attribute with the same name and dimension but on the project
        project_scoped_attribute = get_attribute_by_name_and_dimension(
            name, dimension, project_id=net.project_id)
        if project_scoped_attribute is not None:
            if do_raise is True:
                raise HydraError(
                    f"Unable to add attribute with name '{name}' and dimension '{dimension}' "
                    f"to network '{network_id}' as an "
                    f"attribute with this name and dimension already "
                    f"exists on the project ({net.project_id})")
            else:
                return False


    if project_id is not None:
        globally_scoped_attribute = get_attribute_by_name_and_dimension(name, dimension)
        if globally_scoped_attribute is not None:
            if do_raise is True:
                raise HydraError(
                    f"Unable to add attribute with name '{name}' and dimension '{dimension}' "
                    f"to project '{project_id}' as an "
                    f"attribute with this name and dimension already exists globally")
            else:
                return False

    return True


@required_perms('add_attribute')
def add_attribute(attr, check_existing=True, **kwargs):
    """
    Add a generic attribute, which can then be used in creating
    a resource attribute, and put into a type.

    .. code-block:: python

        (Attr){
            id = 1020
            name = "Test Attr"
            dimension_id = 123
        }

    """
    log.debug("Adding attribute: %s", attr.name)

    user_id = kwargs.get('user_id')

    if check_existing is False:
        attr_i = _add_attribute(attr, user_id=user_id)
        return attr_i

    try:
        attr_qry = db.DBSession.query(Attr).filter(func.lower(Attr.name) == attr.name.lower(),
                                                   Attr.dimension_id == attr.dimension_id)

        # We only need these 2 clauses, as the network ID is the lowest possible level
        # so there is no need for a clause checking for the project id additionally,
        # as the project ID must be the parent of the network ID, so it is redundant
        # to check explicitly
        if attr.network_id is not None and attr.project_id is None:
            # don't just check for attributes scoped to this network but to attributes
            # scoped to it and all parent projects
            network_project_ids = _get_projects_referenced_by_network_id(
                attr.network_id, **kwargs)

            attr_qry = attr_qry.filter(or_(Attr.network_id == attr.network_id,
                                           Attr.project_id.in_(network_project_ids)
                                          ))
        elif attr.project_id is not None:
            # don't just check for attributes scoped to this project, but to all
            # projects in its hierarchy
            project_project_ids = _get_projects_referenced_by_project_id(
                attr.project_id, **kwargs)

            attr_qry = attr_qry.filter(Attr.project_id.in_(project_project_ids))


        attr_i = attr_qry.one()

        attr_i = JSONObject(attr_i)

        log.info("Attr already exists")

    except NoResultFound:
        #set the user ID to 2 here, as this requires admin priviliges. THis is
        #safe to do because this function has already been checked for add_attribute
        #permission from the caller
        attr_i = _add_attribute(attr, user_id=user_id, do_reassign=True)

    return JSONObject(attr_i)

@required_perms('edit_attribute')
def update_attribute(attr, **kwargs):
    """
    Add a generic attribute, which can then be used in creating
    a resource attribute, and put into a type.

    .. code-block:: python

        (Attr){
            id = 1020
            name = "Test Attr"
            dimension_id = 123
        }

    """

    existing_attr_qry = db.DBSession.query(Attr).filter(
        Attr.name == attr.name,
        Attr.dimension_id == attr.dimension_id,
        Attr.id != attr.id)

    if attr.network_id is not None:
        existing_attr_qry = existing_attr_qry.filter(Attr.network_id == attr.network_id)

    if attr.project_id is not None:
        existing_attr_qry = existing_attr_qry.filter(Attr.project_id == attr.project_id)

    existing_attr_i = existing_attr_qry.first()

    if existing_attr_i is not None:
        dimension_name = 'None'
        if attr.dimension_id is not None:
            dimension = units.get_dimension(attr.dimension_id)
            dimension_name = dimension.name
        raise HydraError(f"Cannot update attribute. An attribute with name {attr.name}"
                         f" and dimension {dimension_name} already exists with "
                         f"ID {existing_attr_i.id}")

    log.debug("Updating attribute: %s", attr.name)
    attr_i = _get_attr(attr.id)
    attr_i.name = attr.name
    attr_i.dimension_id = attr.dimension_id
    attr_i.description = attr.description
    attr_i.network_id = attr.network_id
    attr_i.project_id = attr.project_id

    db.DBSession.flush()
    return JSONObject(attr_i)


def delete_attribute(attr_id, **kwargs):
    try:
        attribute = db.DBSession.query(Attr).filter(Attr.id == attr_id).one()
        db.DBSession.delete(attribute)
        db.DBSession.flush()
        return True
    except NoResultFound:
        raise ResourceNotFoundError("Attribute (attribute id=%s) does not exist"%(attr_id))
    except IntegrityError:
        raise HydraError("Unable to delete this attribute as it is in use in a Network")


def _get_projects_referenced_by_network_id(network_id, **kwargs):
    """
        Given a network ID, return the project IDS in which it resides.
        This is used to determine whether there are attributes defined at the project
        level, when trying to add an attribute at the network level.
    """
    #projects in which the networks reside
    project_id_rs = db.DBSession.query(Network.project_id).filter(
        Network.id==network_id).one()
    project_id = project_id_rs.project_id

    network_project_ids = []
    #imported here to avoid a circular import
    from . import project as projectlib
    #we can't just get the project ID. we need to get the whole hierarchy.
    hier = projectlib.get_project_hierarchy(project_id, **kwargs)
    for p in reversed(hier):
        if p.id not in network_project_ids:
            network_project_ids.append(p.id)

    return network_project_ids

def _get_projects_referenced_by_project_id(project_id, **kwargs):
    """
        Given a project ID, return the project IDS in which it resides.
        This is used to determine whether there are attributes defined at the project
        level, when trying to add an attribute at the network level.
    """
    #projects in which the networks reside

    #imported here to avoid a circular import
    from . import project as projectlib
    #we can't just get the project ID. we need to get the whole hierarchy.
    parent_project_ids = []
    hier = projectlib.get_project_hierarchy(project_id, **kwargs)
    for p in reversed(hier):
        if p.id not in parent_project_ids:
            parent_project_ids.append(p.id)

    return parent_project_ids


def add_attributes(attrs, **kwargs):
    """
    Add a list of generic attributes, which can then be used in creating
    a resource attribute, and put into a type.

    .. code-block:: python

        (Attr){
            name = "Test Attr"
            dimension_id (optional) = 123 # the ID of the relevant dimension. Defaults to None.
            project_id (optional) = 123,
            network_id (optional) = 456
        }

    Attributes can only be added to one network / project a a time. Raises a
    Hydra Error if it finds different networks or project IDs in the request.

    """

    #Check to see if any of the attributs being added are already there.
    #If they are there already, don't add a new one. If an attribute
    #with the same name is there already but with a different dimension,
    #add a new attribute.
    user_id = kwargs.get('user_id')

    # All global attributes
    global_attrs = db.DBSession.query(Attr).filter(and_(Attr.network_id == None, Attr.project_id == None)).all()

    #project-scoped attributes
    project_attrs = []

    network_ids = list(set([a.network_id for a in filter(lambda x:x.network_id is not None and x.project_id is None, attrs)]))

    if len(network_ids) > 1:
        raise HydraError("Cannot bulk add attributes to different networks.")

    #if the attributes are specified with a network ID but not a project ID, then find the
    #project IDS for those network so we can check if there are
    #matching attributes scoped to the container projects
    network_project_ids = []
    if len(network_ids) > 0:
        network_project_ids = _get_projects_referenced_by_network_id(network_ids[0], **kwargs)

    #Get all the project IDs specified in the incoming attributes.
    #Attributes can only ne added to one project at a time

    direct_project_ids = list(set([a.project_id for a in filter(lambda x:x.project_id is not None, attrs)]))

    if len(direct_project_ids) > 1:
        raise HydraError("Cannot bulk add attributes to different projects.")

    project_ids = []
    if len(direct_project_ids) > 0:
        project_ids = _get_projects_referenced_by_project_id(direct_project_ids[0], **kwargs)

    #go top down, saving the attributes, and overwriting the duplicates as we
    #go down the tree.
    project_attr_dict = {}
    seen = []
    for project_id in project_ids + network_project_ids:
        #  avoid duplicates. Can't use a set here because order is important
        if project_id in seen:
            continue
        else:
            seen.append(project_id)
        project_attrs = db.DBSession.query(Attr).filter(Attr.project_id == project_id).all()
        for project_attr in project_attrs:
            project_attr_dict[(project_attr.name, project_attr.dimension)] = project_attr

    #network scoped attributes
    network_attrs = []
    if len(network_ids) > 0:
        network_attrs = db.DBSession.query(Attr).filter(Attr.network_id.in_(network_ids)).all()

    all_attrs = global_attrs + list(project_attr_dict.values()) + network_attrs

    #this iterates from the highest scope to the lowst, thus overwriting higher-level
    #attributes with lower level ones.
    attr_dict = {}
    for attr in all_attrs:
        attr_dict[(attr.name.lower(), attr.dimension_id)] = JSONObject(attr)

    attrs_to_add = []
    existing_attrs = []
    for potential_new_attr in attrs:
        if potential_new_attr is not None:
            # If the attrinute is None we cannot manage it
            log.debug("Adding attribute: %s", potential_new_attr)
            key = (potential_new_attr.name.lower(), potential_new_attr.dimension_id)
            if attr_dict.get(key) is not None:
                existing_attrs.append(attr_dict.get(key))
            else:
                attrs_to_add.append(JSONObject(potential_new_attr))

    new_attrs = []
    for attr in attrs_to_add:
        new_attr_i = _add_attribute(attr, flush=True, user_id=user_id)
        new_attrs.append(new_attr_i)

    db.DBSession.flush()

    new_attrs = new_attrs + existing_attrs

    return [JSONObject(a) for a in new_attrs]

def get_attributes(network_id=None,
                   project_id=None,
                   include_global=False,
                   include_network_attributes=False,
                   include_hierarchy=False, **kwargs):
    """
        Get all attributes.
        args:
            network_id (optional): Return network-scoped attributes (attributes defined only on a network)
            project_id (optional): Return project-scoped attributes (attributes defined only on a project)
            include_global (Bool): If a network ID or project ID are specified, global attributes are
                                   not returned unless this flag is True.
            include_network_attributes (Bool): If a project ID is specified but not a network ID, then use
                                               this flag to indicate whether the attributes scoped to all networks
                                               inside the specified project should also be returned.
            include_hierarchy (Bool): Include attributes from projects higher up in the project hierarchy
    """

    base_qry = db.DBSession.query(Attr)

    if (network_id is None and project_id is None) or include_global is True:
        #First get all global attributes
        attrs = base_qry.filter(
            and_(
                Attr.network_id == None,
                Attr.project_id == None
            )
            ).all()

        global_attrs = [JSONObject(a) for a in attrs]
    else:
        global_attrs = []

    project_scoped_attributes = []
    network_scoped_attributes = []


    #Now get all project attributes
    if project_id is not None:
        project = db.DBSession.query(Project).filter(Project.id==project_id).one()
        project_scoped_attributes = project.get_scoped_attributes(include_hierarchy=include_hierarchy)

        if network_id is None and include_network_attributes is True:
            nets = db.DBSession.query(Network).filter(Network.project_id==project_id).all()
            netlookup = {n.id:n for n in nets}
            network_attributes = base_qry.filter(Attr.network_id.in_([n.id for n in nets])).all()
            network_scoped_attributes = [JSONObject(a) for a in network_attributes]
            for nsa in network_scoped_attributes:
                nsa.network_name = netlookup[nsa.network_id].name

    if network_id is not None:
        net = db.DBSession.query(Network).filter(Network.id==network_id).one()
        #don't get the hierarchy if this has already been retrieved by the project
        #attribute retrieval
        include_network_hierarchy=include_hierarchy
        if project_id is not None:
            include_network_hierarchy=False
        network_scoped_attributes = net.get_scoped_attributes(include_hierarchy=include_network_hierarchy)

    all_attrs = network_scoped_attributes + project_scoped_attributes + global_attrs

    all_attrs = sorted(all_attrs, key=lambda x: x.name)

    return all_attrs

def get_attributes_by_name_and_dimension(name, dimension_id=None, **kwargs):
    """
        Get all attributes with the specified name and dimension, irrespective of
        scoping.
        dimension_id can be None, because in attribute the dimension_id is not anymore mandatory
        args:
            name (str): The name of the attribute. Lower() is called on this for comparison, so this
                        is case-insensitive
            dimension_id (int): the ID of the dimension of the attribute
        returns:
            list: JSONObjects derived from the Sqlalchemy rows.
    """
    log.info("Retrieving all attributes with name %s and dimension %s", name, dimension_id)
    attr_qry = db.DBSession.query(Attr).filter(
        and_(
            func.lower(Attr.name) == name.strip().lower(),
            Attr.dimension_id == dimension_id
        )
    )

    attrs_i = attr_qry.all()

    log.info("Found %s attributes", len(attrs_i))

    return attrs_i


def _get_attr(attr_id):
    try:
        attr = db.DBSession.query(Attr).filter(Attr.id == attr_id).one()
        return attr
    except NoResultFound:
        raise ResourceNotFoundError("Attribute with ID %s not found"%(attr_id,))

def _get_templatetype(type_id):
    try:
        typ = db.DBSession.query(TemplateType).filter(TemplateType.id == type_id).one()
        return typ
    except NoResultFound:
        raise ResourceNotFoundError("Template Type with ID %s not found"%(type_id,))

def update_resource_attribute(resource_attr_id, is_var, **kwargs):
    """
        Deletes a resource attribute and all associated data.
    """
    user_id = kwargs.get('user_id')
    try:
        ra = db.DBSession.query(ResourceAttr).filter(ResourceAttr.id == resource_attr_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Resource Attribute %s not found"%(resource_attr_id))

    ra.check_write_permission(user_id)

    ra.is_var = is_var

    return 'OK'

def delete_resource_attribute(resource_attr_id, **kwargs):
    """
        Deletes a resource attribute and all associated data.
    """
    user_id = kwargs.get('user_id')
    try:
        ra = db.DBSession.query(ResourceAttr).filter(ResourceAttr.id == resource_attr_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Resource Attribute %s not found"%(resource_attr_id))

    ra.check_write_permission(user_id)
    db.DBSession.delete(ra)
    db.DBSession.flush()
    return 'OK'

def add_resource_attributes(resource_attributes, **kwargs):
    """
    Adds a list of resource_attributes to the specified resource.

    Returns a dict mapping a tuple containing the (resource_id, attr_id)
    of each argument resource_attribute to the id with which it was
    added.

    A resource attribute looks like:
    {
        "ref_key": "NODE",
        "node_id": 1,
        "link_id": None,
        "group_id": None,
        "network_id": None,
        "attr_id": 1,
        "is_var": "N"}
    """
    if len(resource_attributes) == 0:
        return {}

    #1. Identify the network ID
    network_id = get_network_id_from_resource_attribute(resource_attributes[0])
    #2. Get all the resource attributes in the network
    network_resource_attributes = get_all_network_resourceattributes(network_id, **kwargs)
    #3. Remove any duplicates from the incoming data in case there are RAs which are already there
    network_ra_lookup = {(ra.attr_id, ra.ref_key, _get_resource_id(ra)): ra for ra in network_resource_attributes}

    #an RA in the database has a 'REF_KEY' column, and then a 'network_id', 'node_id', 'link_id', 'group_id' and 'project_id' column which are mutually exclusive.
    #The incoming RA can have this format, but also 'ref_key', and 'ref_id', where ref_id is the ID of the resource, and ref_key is the type of resource.
    #The incoming RA can also have a 'resource_type' and 'resource_id' column, which is the same as the ref_key and ref_id.
    #The result should be a ref_key and the relevant resource_id column (network_id, node_id etc) set to the ID of the resource.
    key_to_field = {
        'NETWORK': 'network_id',
        'NODE': 'node_id',
        'LINK': 'link_id',
        'GROUP': 'group_id',
        'PROJECT': 'project_id',
    }

    field_to_key = {v: k for k, v in key_to_field.items()}

    for ra in resource_attributes:
        if ra.get('is_var') is not None:
            ra['attr_is_var'] = ra['is_var']

        # If no ref_key, try to infer it from existing ID fields
        if ra.get('ref_key') is None:
            for field, key in field_to_key.items():
                if ra.get(field) is not None:
                    ra['ref_key'] = key
                    break

        # Override with resource_type if provided
        if ra.get('resource_type'):
            ra['ref_key'] = ra['resource_type']

        # If resource_id is present and ref_key is known, populate correct field
        if ra.get('resource_id') and ra.get('ref_key'):
            target_field = key_to_field.get(ra['ref_key'])
            if target_field:
                ra[target_field] = ra['resource_id']

        # If ref_key and ref_id are both provided, override and clear other fields
        if ra.get('ref_key') and ra.get('ref_id'):
            # Clear all ID fields
            for field in key_to_field.values():
                ra[field] = None

            # Set the correct ID field
            target_field = key_to_field.get(ra['ref_key'])
            if target_field:
                ra[target_field] = ra['ref_id']

    ras_to_be_inserted = []

    for ra in resource_attributes:
        key = (ra.attr_id, ra['ref_key'], _get_resource_id(ra))
        if key not in network_ra_lookup:
            ras_to_be_inserted.append(ra)

    inserted_ids = {}

    #4. Add the new resource attributes
    cols = list(filter(lambda x: x not in ['id', 'cr_date', 'updated_at'], [c.name for c in ResourceAttr.__table__.columns]))
    ras_to_be_inserted = [{k: v for k, v in ra.items() if k in cols} for ra in ras_to_be_inserted]

    if len(ras_to_be_inserted) > 0:
        log.info("Adding %s new resource attributes", len(ras_to_be_inserted))
        objs = [ResourceAttr(**ra) for ra in ras_to_be_inserted]
        db.DBSession.add_all(objs)
        db.DBSession.flush()  # or commit
        for obj in objs:
            inserted_ids[(obj.get_resource_id(), obj.attr_id)] = obj.id
        # Mark the session as dirty to ensure that the changes are saved
        # This is necessary if you are using a session with autocommit=False
        mark_changed(db.DBSession())

        cache.set(f'network_resource_attributes_{network_id}',
            cache.get(f'network_resource_attributes_{network_id}') + [JSONObject(obj) for obj in objs], 60*60)

    db.DBSession.flush()

    return inserted_ids

def get_network_id_from_resource_attribute(ra):
    network_id = None
    if ra.get('resource_id') is not None:
        #set the node_id, group_id, link_id based on the resource_id
        resource_id = ra.get('resource_id')
        if ra.resource_type == 'NODE':
            ra.node_id = resource_id
        elif ra.resource_type == 'LINK':
            ra.link_id = resource_id
        elif ra.resource_type == 'GROUP':
            ra.group_id = resource_id
        elif ra.resource_type == 'NETWORK':
            ra.network_id = resource_id

    if ra.get('network_id') is not None:
        network_id = ra.get('network_id')
    if ra.get('node_id') is not None:
        node_id = ra.get('node_id')
        node = db.DBSession.query(Node).filter(Node.id == node_id).one()
        network_id = node.network_id
    if ra.get('link_id') is not None:
        link_id = ra.get('link_id')
        link = db.DBSession.query(Link).filter(Link.id == link_id).one()
        network_id = link.network_id
    if ra.get('group_id') is not None:
        group_id = ra.get('group_id')
        group = db.DBSession.query(ResourceGroup).filter(ResourceGroup.id == group_id).one()
        network_id = group.network_id
    return network_id

def add_resource_attribute(resource_type, resource_id, attr_id, is_var, error_on_duplicate=True, **kwargs):
    """
        Add a resource attribute attribute to a resource.

        attr_is_var indicates whether the attribute is a variable or not --
        this is used in simulation to indicate that this value is expected
        to be filled in by the simulator.
    """

    attr = db.DBSession.query(Attr).filter(Attr.id == attr_id).first()

    if attr is None:
        raise ResourceNotFoundError("Attribute with ID %s does not exist."%attr_id)

    resource_i = _get_resource(resource_type, resource_id)

    resourceattr_qry = db.DBSession.query(ResourceAttr).filter(ResourceAttr.ref_key == resource_type)

    if resource_type == 'NETWORK':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.network_id == resource_id)
    elif resource_type == 'NODE':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.node_id == resource_id)
    elif resource_type == 'LINK':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.link_id == resource_id)
    elif resource_type == 'GROUP':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.group_id == resource_id)
    elif resource_type == 'PROJECT':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.project_id == resource_id)
    else:
        raise HydraError('Resource type "{}" not recognised.'.format(resource_type))
    resource_attrs = resourceattr_qry.all()

    for ra in resource_attrs:
        if ra.attr_id == attr_id:
            if not error_on_duplicate:
                return ra

            raise HydraError("Duplicate attribute. %s %s already has attribute %s"
                             %(resource_type, resource_i.get_name(), attr.name))

    attr_is_var = 'Y' if is_var in (True, 'Y') else 'N'

    new_ra = resource_i.add_attribute(attr_id, attr_is_var)
    db.DBSession.flush()

    return new_ra

def add_resource_attrs_from_type(type_id, resource_type, resource_id, **kwargs):
    """
        adds all the attributes defined by a type to a node.
    """
    type_i = _get_templatetype(type_id)

    resource_i = _get_resource(resource_type, resource_id)

    resourceattr_qry = db.DBSession.query(ResourceAttr).filter(ResourceAttr.ref_key == resource_type)

    if resource_type == 'NETWORK':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.network_id == resource_id)
    elif resource_type == 'NODE':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.node_id == resource_id)
    elif resource_type == 'LINK':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.link_id == resource_id)
    elif resource_type == 'GROUP':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.group_id == resource_id)
    elif resource_type == 'PROJECT':
        resourceattr_qry = resourceattr_qry.filter(ResourceAttr.project_id == resource_id)

    resource_attrs = resourceattr_qry.all()

    attrs = {}
    for res_attr in resource_attrs:
        attrs[res_attr.attr_id] = res_attr

    new_resource_attrs = []
    for item in type_i.typeattrs:
        if attrs.get(item.attr_id) is None:
            ra = resource_i.add_attribute(item.attr_id)
            new_resource_attrs.append(ra)

    db.DBSession.flush()

    return new_resource_attrs

def get_all_network_resourceattributes(network_id, template_id=None, return_orm=False, **kwargs):
    """
        Get all the resource attributes for all the nodes, links and groups in the network
        including network attributes. This is used primarily to avoid retrieving
        all global attributes for menus etc, most of which are not necessary.

        args:
            network_id (int): The ID of the network containing the attributes
            template_id (int): A filter which will cause the function to
                                return attributes associated to that template
            return_orm (bool): Flag to force the function to return ORM objects instead
                                of JSONObjects, likely to be used internally from another
                                function

        returns:
            A list of Attributes as JSONObjects, with the
            additional data of 'attr_is_var'
            from its assocated ResourceAttribute. ex:
                {id:123,
                name: 'cost'
                dimension_id: 124,
                attr_is_var: 'Y' #comes from the ResourceAttr
                }
    """

    user_id = kwargs.get('user_id')
    net = _get_network(network_id)
    net.check_read_permission(user_id, do_raise=True)

    if cache.get(f'network_resource_attributes_{network_id}') is not None:
        return cache.get(f'network_resource_attributes_{network_id}')

    network_attrs = db.DBSession.query(ResourceAttr).\
            join(Attr, ResourceAttr.attr_id == Attr.id).\
            join(Network, Network.id == ResourceAttr.network_id).\
            filter(ResourceAttr.network_id != None).\
            filter(Network.id==network_id).all()

    node_attrs = db.DBSession.query(ResourceAttr).\
            join(Attr, ResourceAttr.attr_id == Attr.id).\
            join(Node, Node.id == ResourceAttr.node_id).\
            join(Network, Network.id == Node.network_id).\
            filter(ResourceAttr.node_id != None).\
            filter(Network.id==network_id).all()

    link_attrs = db.DBSession.query(ResourceAttr).\
            join(Attr, ResourceAttr.attr_id == Attr.id).\
            join(Link, Link.id == ResourceAttr.link_id).\
            join(Network, Network.id == Link.network_id).\
            filter(ResourceAttr.link_id != None).\
            filter(Network.id==network_id).all()

    group_attrs = db.DBSession.query(ResourceAttr).\
            join(Attr, ResourceAttr.attr_id == Attr.id).\
            join(ResourceGroup, ResourceGroup.id == ResourceAttr.group_id).\
            join(Network, Network.id == ResourceGroup.network_id).\
            filter(ResourceAttr.group_id != None).\
            filter(Network.id==network_id).all()

    resource_attrs = network_attrs + node_attrs + link_attrs + group_attrs

    if template_id is not None:
        attr_ids = []
        rs = db.DBSession.query(TypeAttr).join(
            TemplateType,
            TemplateType.id == TypeAttr.type_id).filter(
            TemplateType.template_id == template_id).all()

        for r in rs:
            attr_ids.append(r.attr_id)

        resource_attrs = filter(lambda x: x.attr_id in attr_ids, resource_attrs)

    network_attributes = []
    for ra in resource_attrs:
        if return_orm is True:
            network_attributes.append(ra)
        else:
            ra_j = JSONObject(ra)
            ra_j.attr = JSONObject(ra.attr)
            network_attributes.append(ra_j)

    cache.set(f'network_resource_attributes_{network_id}', network_attributes, 60*60)

    return network_attributes


def get_all_network_attributes(network_id, template_id=None, **kwargs):
    """
        Get all the attributes for all the nodes, links and groups in the network
        including network attributes. This is used primarily to avoid retrieving
        all global attributes for menus etc, most of which are not necessary.

        args:
            network_id (int): The ID of the network containing the attributes
            template_id (int): A filter which will cause the function to
                                return attributes associated to that template

        returns:
            A list of Attributes as JSONObjects, with the
            additional data of 'attr_is_var'
            from its assocated ResourceAttribute. ex:
                {id:123,
                name: 'cost'
                dimension_id: 124,
                attr_is_var: 'Y' #comes from the ResourceAttr
                }
        NOTE: This was originally done with a single query, but was split up for
              performamce reasons
    """
    user_id = kwargs.get('user_id')
    net = _get_network(network_id)
    net.check_read_permission(user_id, do_raise=True)

    network_attr_qry = db.DBSession.query(Attr, ResourceAttr.attr_is_var).\
            join(ResourceAttr, ResourceAttr.attr_id == Attr.id).\
            join(Network, Network.id == ResourceAttr.network_id).filter(
            and_(ResourceAttr.network_id != None,
                 ResourceAttr.network_id == network_id))
    network_attrs = network_attr_qry.all()

    node_attr_qry = db.DBSession.query(Attr, ResourceAttr.attr_is_var).\
            join(ResourceAttr, ResourceAttr.attr_id == Attr.id).\
            join(Node, Node.id == ResourceAttr.node_id).filter(
                and_(ResourceAttr.node_id is not None,
                 ResourceAttr.node_id == Node.id,
                 Node.network_id == network_id))
    node_attrs = node_attr_qry.all()

    link_attr_qry = db.DBSession.query(Attr, ResourceAttr.attr_is_var).\
            join(ResourceAttr, ResourceAttr.attr_id==Attr.id).\
            join(Link, Link.id == ResourceAttr.link_id).filter(
                and_(ResourceAttr.link_id is not None,
                 ResourceAttr.link_id == Link.id,
                 Link.network_id==network_id))
    link_attrs = link_attr_qry.all()

    group_attr_qry = db.DBSession.query(Attr, ResourceAttr.attr_is_var).\
            join(ResourceAttr, ResourceAttr.attr_id == Attr.id).\
            join(ResourceGroup, ResourceGroup.id == ResourceAttr.group_id).filter(
                and_(ResourceAttr.group_id is not None,
                 ResourceAttr.group_id == ResourceGroup.id,
                 ResourceGroup.network_id == network_id))

    group_attrs = group_attr_qry.all()

    resource_attrs = network_attrs + node_attrs + link_attrs + group_attrs

    if template_id is not None:
        log.info("Filtering out only attributes which appear in template %s", template_id)
        attr_ids = []
        rs = db.DBSession.query(TypeAttr).join(
            TemplateType,
            TemplateType.id == TypeAttr.type_id).filter(
            TemplateType.template_id == template_id).all()

        for r in rs:
            attr_ids.append(r.attr_id)
        filtered_results = []
        for ra in resource_attrs:
            if ra[0].id in attr_ids:
                filtered_results.append(ra)

        log.info("Filtered out %s attributes", len(resource_attrs)-len(filtered_results))

        resource_attrs = filtered_results


    network_attributes = []
    for ra in resource_attrs:
        attr_j = JSONObject(ra[0])
        attr_j.attr_is_var = ra[1]
        network_attributes.append(attr_j)

    return network_attributes

@required_perms('get_network')
def get_all_resource_attributes(ref_key, network_id, template_id=None, **kwargs):
    """
        Get all the resource attributes for a given resource type in the network.
        That includes all the resource attributes for a given type within the network.
        For example, if the ref_key is 'NODE', then it will return all the attirbutes
        of all nodes in the network. This function allows a front end to pre-load an entire
        network's resource attribute information to reduce on function calls.
        If type_id is specified, only
        return the resource attributes within the type.
    """

    user_id = kwargs.get('user_id')

    net = _get_network(network_id)
    net.check_read_permission(user_id, do_raise=True)

    resource_attr_qry = db.DBSession.query(ResourceAttr).\
        outerjoin(Node, Node.id == ResourceAttr.node_id).\
        outerjoin(Link, Link.id == ResourceAttr.link_id).\
        outerjoin(ResourceGroup, ResourceGroup.id == ResourceAttr.group_id).filter(
            ResourceAttr.ref_key == ref_key,
            or_(
                and_(ResourceAttr.node_id != None,
                     ResourceAttr.node_id == Node.id,
                     Node.network_id == network_id),

                and_(ResourceAttr.link_id != None,
                     ResourceAttr.link_id == Link.id,
                     Link.network_id == network_id),

                and_(ResourceAttr.group_id != None,
                     ResourceAttr.group_id == ResourceGroup.id,
                     ResourceGroup.network_id == network_id)
            ))

    if template_id is not None:
        attr_ids = []
        rs = db.DBSession.query(TypeAttr).join(
            TemplateType,
            TemplateType.id == TypeAttr.type_id).filter(
                TemplateType.template_id == template_id).all()
        for r in rs:
            attr_ids.append(r.attr_id)

        resource_attr_qry = resource_attr_qry.filter(ResourceAttr.attr_id.in_(attr_ids))

    resource_attrs = resource_attr_qry.all()

    return resource_attrs

def get_resource_attributes(ref_key, ref_id, type_id=None, **kwargs):
    """
        Get all the resource attributes for a given resource.
        If type_id is specified, only
        return the resource attributes within the type.
    """

    user_id = kwargs.get('user_id')

    resource_attr_qry = db.DBSession.query(ResourceAttr).filter(
        ResourceAttr.ref_key == ref_key,
        or_(
            ResourceAttr.network_id == ref_id,
            ResourceAttr.node_id == ref_id,
            ResourceAttr.link_id == ref_id,
            ResourceAttr.group_id == ref_id
        ))

    if type_id is not None:
        attr_ids = []
        rs = db.DBSession.query(TypeAttr).filter(TypeAttr.type_id == type_id).all()
        for r in rs:
            attr_ids.append(r.attr_id)

        resource_attr_qry = resource_attr_qry.filter(ResourceAttr.attr_id.in_(attr_ids))

    resource_attrs = resource_attr_qry.all()

    return resource_attrs

def check_attr_dimension(attr_id, **kwargs):
    """
        Check that the dimension of the resource attribute data is consistent
        with the definition of the attribute.
        If the attribute says 'volume', make sure every dataset connected
        with this attribute via a resource attribute also has a dimension
        of 'volume'.
    """
    attr_i = _get_attr(attr_id)

    datasets = db.DBSession.query(Dataset).filter(
        Dataset.id == ResourceScenario.dataset_id,
        ResourceScenario.resource_attr_id == ResourceAttr.id,
        ResourceAttr.attr_id == attr_id).all()

    bad_datasets = []
    for d in datasets:
        if  attr_i.dimension_id is None and d.unit is not None or \
            attr_i.dimension_id is not None and d.unit is None or \
            units.get_dimension_by_unit_id(d.unit_id) != attr_i.dimension_id:
                # If there is an inconsistency
            bad_datasets.append(d.id)

    if len(bad_datasets) > 0:
        raise HydraError("Datasets %s have a different dimension_id to attribute %s"%(bad_datasets, attr_id))

    return 'OK'

def get_resource_attribute(resource_attr_id, **kwargs):
    """
        Get a specific resource attribte, by ID
        If type_id is Gspecified, only
        return the resource attributes within the type.
    """
    user_id = kwargs.get('user_id')

    resource_attr_qry = db.DBSession.query(ResourceAttr).filter(
        ResourceAttr.id == resource_attr_id,
    )

    resource_attr = resource_attr_qry.first()

    resource = resource_attr.get_resource()
    if resource_attr.network_id is not None:
        network = resource
    else:
        network = resource.network

    network.check_read_permission(user_id)

    if resource_attr is None:
        raise ResourceNotFoundError(f"Resource attribute {resource_attr_id} does not exist")

    return resource_attr

def set_attribute_mapping(resource_attr_a, resource_attr_b, **kwargs):
    """
        Define one resource attribute from one network as being the same as
        that from another network.
    """
    user_id = kwargs.get('user_id')
    ra_1 = get_resource_attribute(resource_attr_a, user_id=user_id)
    ra_2 = get_resource_attribute(resource_attr_b, user_id=user_id)

    mapping = ResourceAttrMap(
        resource_attr_id_a=resource_attr_a,
        resource_attr_id_b=resource_attr_b,
        network_a_id=ra_1.get_network().id,
        network_b_id=ra_2.get_network().id
    )

    db.DBSession.add(mapping)

    db.DBSession.flush()

    return mapping

def delete_attribute_mapping(resource_attr_a, resource_attr_b, **kwargs):
    """
        Define one resource attribute from one network as being the same as
        that from another network.
    """

    rm = aliased(ResourceAttrMap, name='rm')

    log.info("Trying to delete attribute map. %s -> %s", resource_attr_a, resource_attr_b)
    mapping = db.DBSession.query(rm).filter(
        rm.resource_attr_id_a == resource_attr_a,
        rm.resource_attr_id_b == resource_attr_b).first()

    if mapping is not None:
        log.info("Deleting attribute map. %s -> %s", resource_attr_a, resource_attr_b)
        db.DBSession.delete(mapping)
        db.DBSession.flush()

    return 'OK'

def delete_mappings_in_network(network_id, network_2_id=None, **kwargs):
    """
        Delete all the resource attribute mappings in a network. If another network
        is specified, only delete the mappings between the two networks.
    """
    user_id = kwargs.get('user_id')
    net = _get_network(network_id)
    net.check_read_permission(user_id, do_raise=True)

    qry = db.DBSession.query(ResourceAttrMap).filter(or_(ResourceAttrMap.network_a_id == network_id, ResourceAttrMap.network_b_id == network_id))

    if network_2_id is not None:
        qry = qry.filter(or_(ResourceAttrMap.network_a_id==network_2_id, ResourceAttrMap.network_b_id==network_2_id))

    mappings = qry.all()

    for m in mappings:
        db.DBSession.delete(m)
    db.DBSession.flush()

    return 'OK'

def get_mappings_in_network(network_id, network_2_id=None, **kwargs):
    """
        Get all the resource attribute mappings in a network. If another network
        is specified, only return the mappings between the two networks.
    """
    qry = db.DBSession.query(ResourceAttrMap).filter(or_(ResourceAttrMap.network_a_id == network_id, ResourceAttrMap.network_b_id == network_id))

    if network_2_id is not None:
        qry = qry.filter(or_(ResourceAttrMap.network_a_id==network_2_id, ResourceAttrMap.network_b_id==network_2_id))

    return qry.all()

def get_node_mappings(node_id, node_2_id=None, **kwargs):
    """
        Get all the resource attribute mappings in a network. If another network
        is specified, only return the mappings between the two networks.
    """
    qry = db.DBSession.query(ResourceAttrMap).filter(
        or_(
            and_(
                ResourceAttrMap.resource_attr_id_a == ResourceAttr.id,
                ResourceAttr.node_id == node_id),
            and_(
                ResourceAttrMap.resource_attr_id_b == ResourceAttr.id,
                ResourceAttr.node_id == node_id)))

    if node_2_id is not None:
        aliased_ra = aliased(ResourceAttr, name="ra2")
        qry = qry.filter(or_(
            and_(
                ResourceAttrMap.resource_attr_id_a == aliased_ra.id,
                aliased_ra.node_id == node_2_id),
            and_(
                ResourceAttrMap.resource_attr_id_b == aliased_ra.id,
                aliased_ra.node_id == node_2_id)))

    return qry.all()

def get_link_mappings(link_id, link_2_id=None, **kwargs):
    """
        Get all the resource attribute mappings in a network. If another network
        is specified, only return the mappings between the two networks.
    """
    qry = db.DBSession.query(ResourceAttrMap).filter(
        or_(
            and_(
                ResourceAttrMap.resource_attr_id_a == ResourceAttr.id,
                ResourceAttr.link_id == link_id),
            and_(
                ResourceAttrMap.resource_attr_id_b == ResourceAttr.id,
                ResourceAttr.link_id == link_id)))

    if link_2_id is not None:
        aliased_ra = aliased(ResourceAttr, name="ra2")
        qry = qry.filter(or_(
            and_(
                ResourceAttrMap.resource_attr_id_a == aliased_ra.id,
                aliased_ra.link_id == link_2_id),
            and_(
                ResourceAttrMap.resource_attr_id_b == aliased_ra.id,
                aliased_ra.link_id == link_2_id)))

    return qry.all()


def get_network_mappings(network_id, network_2_id=None, **kwargs):
    """
        Get all the mappings of network resource attributes, NOT ALL THE MAPPINGS
        WITHIN A NETWORK. For that, ``use get_mappings_in_network``. If another network
        is specified, only return the mappings between the two networks.
    """
    qry = db.DBSession.query(ResourceAttrMap).filter(
        or_(
            and_(
                ResourceAttrMap.resource_attr_id_a == ResourceAttr.id,
                ResourceAttr.network_id == network_id),
            and_(
                ResourceAttrMap.resource_attr_id_b == ResourceAttr.id,
                ResourceAttr.network_id == network_id)))

    if network_2_id is not None:
        aliased_ra = aliased(ResourceAttr, name="ra2")
        qry = qry.filter(or_(
            and_(
                ResourceAttrMap.resource_attr_id_a == aliased_ra.id,
                aliased_ra.network_id == network_2_id),
            and_(
                ResourceAttrMap.resource_attr_id_b == aliased_ra.id,
                aliased_ra.network_id == network_2_id)))

    return qry.all()

def check_attribute_mapping_exists(resource_attr_id_source, resource_attr_id_target, **kwargs):
    """
        Check whether an attribute mapping exists between a source and target resource attribute.
        returns 'Y' if a mapping exists. Returns 'N' in all other cases.
    """
    qry = db.DBSession.query(ResourceAttrMap).filter(
                ResourceAttrMap.resource_attr_id_a == resource_attr_id_source,
                ResourceAttrMap.resource_attr_id_b == resource_attr_id_target).all()

    if len(qry) > 0:
        return 'Y'
    else:
        return 'N'


def get_attribute_group(group_id, **kwargs):
    """
        Get a specific attribute group
    """

    user_id=kwargs.get('user_id')

    try:
        group_i = db.DBSession.query(AttrGroup).filter(
                                            AttrGroup.id==group_id).one()
        group_i.project.check_read_permission(user_id)
    except NoResultFound:
        raise HydraError("Group %s not found" % (group_id,))

    return group_i


def add_attribute_group(attributegroup, **kwargs):
    """
        Add a new attribute group.

        An attribute group is a container for attributes which need to be grouped
        in some logical way. For example, if the 'attr_is_var' flag isn't expressive
        enough to delineate different groupings.

        an attribute group looks like:
            {
                'project_id' : XXX,
                'name'       : 'my group name'
                'description : 'my group description' (optional)
                'layout'     : 'my group layout'      (optional)
                'exclusive'  : 'N' (or 'Y' )          (optional, default to 'N')
            }
    """
    log.info("attributegroup.project_id %s",attributegroup.project_id) # It is None while it should be valued
    user_id = kwargs.get('user_id')
    project_i = _get_project(attributegroup.project_id)
    project_i.check_write_permission(user_id)
    try:

        group_i = db.DBSession.query(AttrGroup).filter(
                                            AttrGroup.name==attributegroup.name,
                                            AttrGroup.project_id==attributegroup.project_id).one()
        log.info("Group %s already exists in project %s", attributegroup.name, attributegroup.project_id)

    except NoResultFound:

        group_i = AttrGroup()
        group_i.project_id  = attributegroup.project_id
        group_i.name        = attributegroup.name
        group_i.description = attributegroup.description
        group_i.layout      = attributegroup.get_layout()
        group_i.exclusive   = attributegroup.exclusive

        db.DBSession.add(group_i)
        db.DBSession.flush()

        log.info("Attribute Group %s added to project %s", attributegroup.name, attributegroup.project_id)

    return group_i

def update_attribute_group(attributegroup, **kwargs):
    """
        Add a new attribute group.

        An attribute group is a container for attributes which need to be grouped
        in some logical way. For example, if the 'attr_is_var' flag isn't expressive
        enough to delineate different groupings.

        an attribute group looks like:
            {
                'project_id' : XXX,
                'name'       : 'my group name'
                'description : 'my group description' (optional)
                'layout'     : 'my group layout'      (optional)
                'exclusive'  : 'N' (or 'Y' )          (optional, default to 'N')
            }
    """
    user_id=kwargs.get('user_id')

    if attributegroup.id is None:
        raise HydraError("cannot update attribute group. no ID specified")

    try:

        group_i = db.DBSession.query(AttrGroup).filter(AttrGroup.id==attributegroup.id).one()
        group_i.project.check_write_permission(user_id)

        layout = attributegroup.layout
        group_i.name        = attributegroup.name
        group_i.description = attributegroup.description
        group_i.layout = json.dumps(layout) if not isinstance(layout, str) else layout
        group_i.exclusive   = attributegroup.exclusive

        db.DBSession.flush()

        log.info("Group %s in project %s updated", attributegroup.id, attributegroup.project_id)
    except NoResultFound:

        raise HydraError('No Attribute Group %s was found in project %s', attributegroup.id, attributegroup.project_id)


    return group_i

def delete_attribute_group(group_id, **kwargs):
    """
        Delete an attribute group.
    """
    user_id = kwargs['user_id']

    try:

        group_i = db.DBSession.query(AttrGroup).filter(AttrGroup.id==group_id).one()

        group_i.project.check_write_permission(user_id)

        db.DBSession.delete(group_i)
        db.DBSession.flush()

        log.info("Group %s in project %s deleted", group_i.id, group_i.project_id)
    except NoResultFound:

        raise HydraError('No Attribute Group %s was found', group_id)


    return 'OK'

def get_network_attributegroup_items(network_id, **kwargs):
    """
        Get all the group items in a network
    """

    user_id=kwargs.get('user_id')


    net_i = _get_network(network_id)

    net_i.check_read_permission(user_id)

    group_items_i = db.DBSession.query(AttrGroupItem).filter(
                                    AttrGroupItem.network_id==network_id).all()

    return group_items_i

def get_group_attributegroup_items(network_id, group_id, **kwargs):
    """
        Get all the items in a specified group, within a network
    """
    user_id=kwargs.get('user_id')

    network_i = _get_network(network_id)

    network_i.check_read_permission(user_id)

    group_items_i = db.DBSession.query(AttrGroupItem).filter(
                                    AttrGroupItem.network_id==network_id,
                                    AttrGroupItem.group_id==group_id).all()

    return group_items_i


def get_attribute_item_groups(network_id, attr_id, **kwargs):
    """
        Get all the group items in a network with a given attribute_id
    """
    user_id=kwargs.get('user_id')

    network_i = _get_network(network_id)

    network_i.check_read_permission(user_id)

    group_items_i = db.DBSession.query(AttrGroupItem).filter(
                                        AttrGroupItem.network_id==network_id,
                                        AttrGroupItem.attr_id==attr_id).all()

    return group_items_i

def _get_attr_group(group_id):
    try:
        group_i = db.DBSession.query(AttrGroup).filter(AttrGroup.id==group_id).one()
    except NoResultFound:
        raise HydraError("Error adding attribute group item: group %s not found" % (group_id))

    return group_i

def _get_attributegroupitems(network_id):
    existing_agis = db.DBSession.query(AttrGroupItem).filter(AttrGroupItem.network_id==network_id).all()
    return existing_agis

def add_attribute_group_items(attributegroupitems, **kwargs):

    """
        Populate attribute groups with items.
        ** attributegroupitems : a list of items, of the form:
            ```{
                    'attr_id'    : X,
                    'group_id'   : Y,
                    'network_id' : Z,
               }```

        Note that this approach supports the possibility of populating groups
        within multiple networks at the same time.

        When adding a group item, the function checks whether it can be added,
        based on the 'exclusivity' setup of the groups -- if a group is specified
        as being 'exclusive', then any attributes within that group cannot appear
        in any other group (within a network).
    """

    user_id = kwargs.get('user_id')

    if not isinstance(attributegroupitems, list):
        raise HydraError("Cannpt add attribute group items. Attributegroupitems must be a list")

    new_agis_i = []

    group_lookup = {}

    #for each network, keep track of what attributes are contained in which groups it's in
    #structure: {NETWORK_ID : {ATTR_ID: [GROUP_ID]}
    agi_lookup = {}

    network_lookup = {}

    #'agi' = shorthand for 'attribute group item'
    for agi in attributegroupitems:

        network_i = network_lookup.get(agi.network_id)

        if network_i is None:
            network_i = _get_network(agi.network_id)
            network_lookup[agi.network_id] = network_i

        network_i.check_write_permission(user_id)


        #Get the group so we can check for exclusivity constraints
        group_i = group_lookup.get(agi.group_id)
        if group_i is None:
            group_lookup[agi.group_id] = _get_attr_group(agi.group_id)

        network_agis = agi_lookup

        #Create a map of all agis currently in the network
        if agi_lookup.get(agi.network_id) is None:
            agi_lookup[agi.network_id] = {}
            network_agis = _get_attributegroupitems(agi.network_id)
            log.info(network_agis)
            for net_agi in network_agis:

                if net_agi.group_id not in group_lookup:
                    group_lookup[net_agi.group_id] = _get_attr_group(net_agi.group_id)

                if agi_lookup.get(net_agi.network_id) is None:
                    agi_lookup[net_agi.network_id][net_agi.attr_id] = [net_agi.group_id]
                else:
                    if agi_lookup[net_agi.network_id].get(net_agi.attr_id) is None:
                        agi_lookup[net_agi.network_id][net_agi.attr_id] = [net_agi.group_id]
                    elif net_agi.group_id not in agi_lookup[net_agi.network_id][net_agi.attr_id]:
                        agi_lookup[net_agi.network_id][net_agi.attr_id].append(net_agi.group_id)
        #Does this agi exist anywhere else inside this network?
        #Go through all the groups that this attr is in and make sure it's not exclusive
        if agi_lookup[agi.network_id].get(agi.attr_id) is not None:
            for group_id in agi_lookup[agi.network_id][agi.attr_id]:
                group = group_lookup[group_id]
                #Another group has been found.
                if group.exclusive == 'Y':
                    #The other group is exclusive, so this attr can't be added
                    raise HydraError("Attribute %s is already in Group %s for network %s. This group is exclusive, so attr %s cannot exist in another group."%(agi.attr_id, group.id, agi.network_id, agi.attr_id))

            #Now check that if this group is exclusive, then the attr isn't in
            #any other groups
            if group_lookup[agi.group_id].exclusive == 'Y':
                if len(agi_lookup[agi.network_id][agi.attr_id]) > 0:
                    #The other group is exclusive, so this attr can't be added
                    raise HydraError("Cannot add attribute %s to group %s. This group is exclusive, but attr %s has been found in other groups (%s)" % (agi.attr_id, agi.group_id, agi.attr_id, agi_lookup[agi.network_id][agi.attr_id]))


        agi_i = AttrGroupItem()
        agi_i.network_id = agi.network_id
        agi_i.group_id   = agi.group_id
        agi_i.attr_id    = agi.attr_id

        #Update the lookup table in preparation for the next pass.
        if agi_lookup[agi.network_id].get(agi.attr_id) is None:
            agi_lookup[agi.network_id][agi.attr_id] = [agi.group_id]
        elif agi.group_id not in agi_lookup[agi.network_id][agi.attr_id]:
            agi_lookup[agi.network_id][agi.attr_id].append(agi.group_id)


        db.DBSession.add(agi_i)

        new_agis_i.append(agi_i)
    log.info(agi_lookup)

    db.DBSession.flush()

    return new_agis_i

def delete_attribute_group_items(attributegroupitems, **kwargs):

    """
        remove attribute groups items .
        ** attributegroupitems : a list of items, of the form:
            ```{
                    'attr_id'    : X,
                    'group_id'   : Y,
                    'network_id' : Z,
               }```
    """


    user_id=kwargs.get('user_id')

    log.info("Deleting %s attribute group items", len(attributegroupitems))

    #if there area attributegroupitems from different networks, keep track of those
    #networks to ensure the user actually has permission to remove the items.
    network_lookup = {}

    if not isinstance(attributegroupitems, list):
        raise HydraError("Cannpt add attribute group items. Attributegroupitems must be a list")

    #'agi' = shorthand for 'attribute group item'
    for agi in attributegroupitems:

        network_i = network_lookup.get(agi.network_id)

        if network_i is None:
            network_i = _get_network(agi.network_id)
            network_lookup[agi.network_id] = network_i

        network_i.check_write_permission(user_id)

        agi_i = db.DBSession.query(AttrGroupItem).filter(AttrGroupItem.network_id == agi.network_id,
                                                 AttrGroupItem.group_id == agi.group_id,
                                                 AttrGroupItem.attr_id  == agi.attr_id).first()

        if agi_i is not None:
            db.DBSession.delete(agi_i)

    db.DBSession.flush()

    log.info("Attribute group items deleted")

    return 'OK'

@required_role('admin')
def delete_all_duplicate_attributes(**kwargs):
    """
        duplicate attributes can appear in the DB when attributes are added
        with a dimension of None (because mysql allows multiple entries
        even if there is a unique constraint where one of the values is null)

        This identifies one attribute of a duplicate set and then remaps all pointers to duplicates
        to that attribute, before deleting all other duplicate attributes.

        steps are:
            1: Identify all duplicate attributes
            2: Select one of the duplicates to be the one to keep
            3: Remap all resource attributes and type attributes to point from
               duplicate attrs to the keeper.
            4: Delete the duplicates.
    """

    #step 1: get all attributes and filter out the duplicates
    all_attributes = db.DBSession.query(Attr).all()

    #a lookup based on name/dimension (The apparent unique identifier for an attribute)
    attribute_lookup = defaultdict(lambda: [])
    for attribute in all_attributes:
        key = (attribute.name, attribute.dimension_id)
        attribute_lookup[key].append(attribute)

    #Now identify the dupes -- any of the dict's values which has a length > 1
    duplicate_attributes = filter(lambda x: len(x) > 1, attribute_lookup.values())

    for dupe_list in duplicate_attributes:
        log.info("Duplicate attributes found: name: %s, dimension: %s",
                 dupe_list[0].name, dupe_list[0].dimension_id)
        delete_duplicate_attributes(dupe_list)

    db.DBSession.flush()


@required_perms('delete_attribute', 'edit_network')
def delete_duplicate_resourceattributes(network_id=None, **kwargs):
    """
    for every resource, find any situations where there are duplicate attribute
    names, ex 2 max_flows, but where the attribute IDs are different. In this case,
    remove one of them, and keep the one which is used in the template for that node.
    """

    if network_id is None:
        #get all the resource attrs in the system -- but limit by only inputs
        all_ras = db.DBSession.query(ResourceAttr)\
            .filter(ResourceAttr.attr_is_var == 'N')\
            .options(joinedload(ResourceAttr.attr)).all()
    else:
        all_ras = get_all_network_resourceattributes(network_id, return_orm=True, **kwargs)

    #create a mapping for a node's resource attrs by its ID and the name of the attr
    ra_lookup = defaultdict(lambda: [])
    for ra in all_ras:
        key = (ra.ref_key, ra.get_resource_id(), ra.attr.name)
        ra_lookup[key].append(ra)

    duplicate_ra_list = filter(lambda x: len(x) > 1, ra_lookup.values())

    #we now have the duplicate resourceattrs. We need to identify which of them
    #to delete.
    for duplicate_ras in duplicate_ra_list:
        #first get the type of the resource
        resource = duplicate_ras[0].get_resource()
        #now get all the typeattrs defined for that resource
        resource_typeattrs = []
        for rt in resource.types:
            for ta in rt.get_templatetype().typeattrs:
                resource_typeattrs.append(ta.attr_id)

        data_to_transfer = {}
        ras_to_delete = []
        #now identify the attributes which are not defined in a typeattr, and
        #mark them for deletion
        for duplicate_ra in duplicate_ras:
            if duplicate_ra.attr_id not in resource_typeattrs:

                #we've found  a resource attribute not defined on the type.
                #Now check if there's data associated to it, and not the other.

                ra_rs = db.DBSession.query(ResourceScenario)\
                        .filter(ResourceScenario.resource_attr_id == duplicate_ra.id).first()

                #If this is the case, then remap the resource scenario to the RA
                #we wish to keep.
                if ra_rs is not None:
                    data_to_transfer[ra.attr.name] = ra_rs

                ras_to_delete.append(duplicate_ra)

        #None of the dupes are associated to a template, then delete any with
        #no data. Leave any with data, and let the user deal with them individually
        if len(ras_to_delete) == len(duplicate_ras):
            for ra_to_delete in ras_to_delete:
                ra_rs = db.DBSession.query(ResourceScenario)\
                        .filter(ResourceScenario.resource_attr_id == ra_to_delete.id).first()
                if ra_rs is None:
                    log.info("A duplicate found for %s on %s %s. Deleting as it has no data.",
                             ra_to_delete.attr.name,
                             ra_to_delete.ref_key,
                             resource.name)
                    db.DBSession.delete(ra_to_delete)
                else:
                    log.info("A duplicate found for %s on %s %s. Not deleting as it has data.",
                             ra_to_delete.attr.name,
                             ra_to_delete.ref_key,
                             resource.name)

            #no need to go further, so continue the outer loop
            continue

        #do a second pass, this time transferring data from the bad resourceattributes
        #to the keepers
        for duplicate_ra in duplicate_ras:
            #focus now on the keepers i.e. the ones defined in the template
            if duplicate_ra.attr_id in resource_typeattrs:
                ra_rs = db.DBSession.query(ResourceScenario)\
                    .filter(ResourceScenario.resource_attr_id == duplicate_ra.id).first()

                dupes_rs = data_to_transfer.get(ra.attr.name)
                if ra_rs is None:
                    #if this has no value, see if the dupe has a value and transfer it
                    if dupes_rs is not None:
                        log.info("Updating resource scenario with new reference to %s",
                                 duplicate_ra.attr.name)
                        newrs = ResourceScenario()
                        newrs.scenario_id = dupes_rs.scenario_id
                        newrs.resource_attr_id = duplicate_ra.id
                        newrs.dataset_id = dupes_rs.dataset_id
                        db.DBSession.add(newrs)
                else:
                    #delete any RS which are not needed
                    if dupes_rs is not None:
                        db.DBSession.delete(dupes_rs)


        #now finally delete all the dupes
        for ra_to_delete in ras_to_delete:

            log.info("Deleting resource attr %s (id: %s) from %s %s",\
                     duplicate_ra.attr.name,\
                     duplicate_ra.id,\
                     duplicate_ra.ref_key,\
                     resource.name)
            db.DBSession.delete(ra_to_delete)


def delete_duplicate_attributes(dupe_list):
    """
        Take a list of duplicate attributes and delete all but one.
        Steps are:
            1: Select one of the duplicates to keep
            2: Remap all resource attributes and type attributes to point
               from duplicate attrs to the keeper
            3: Delete the duplicates
        args:
            dupe_list: a list of iterables containing duplicate attributes
    """

    #sort by ID, and choose the attribute with the smallest ID to keep
    #The rationale being that this likely has the most existing references
    dupe_list.sort(key=lambda x: x.id)

    keeper = dupe_list[0]

    #remove all the rest in the list
    attrs_to_remove = dupe_list[1:]

    #remap all the attributes from those to delete to the keeper so they can be removed safely
    for attr_to_remove in attrs_to_remove:
        remap_attribute_reference(attr_to_remove.id, keeper.id)

        #now that the remapping is done, we can safely delete the old attribute
        delete_attribute(attr_to_remove.id)

    db.DBSession.flush()

    return keeper

def remap_attribute_reference(old_attr_id, new_attr_id, flush=False):
    """
        Remap everything which references old_attr_id to reference
        new_attr_id
    """
    #first, remap all the resource attributes
    ras_to_remap = db.DBSession.query(ResourceAttr)\
        .filter(ResourceAttr.attr_id == old_attr_id).all()

    deleted_ras = []
    for ra_to_remap in ras_to_remap:
        #is there an RA with the new attribute already on the same resource?
        existing_ra = db.DBSession.query(ResourceAttr)\
            .filter(ResourceAttr.attr_id == new_attr_id,
                    ResourceAttr.ref_key == ra_to_remap.ref_key,
                    ResourceAttr.node_id == ra_to_remap.node_id,
                    ResourceAttr.link_id == ra_to_remap.link_id,
                    ResourceAttr.group_id == ra_to_remap.group_id,
                    ResourceAttr.network_id == ra_to_remap.network_id).first()

        #if so, then it must be deleted so there is only one on the resource.
        if existing_ra is not None:
            #if the RA to be deleted is linked with data and the one to keep
            #is not, then remap the resource scenario.
            old_ra_rs = db.DBSession.query(ResourceScenario)\
                .filter(ResourceScenario.resource_attr_id == ra_to_remap.id).first()
            new_ra_rs = db.DBSession.query(ResourceScenario)\
                .filter(ResourceScenario.resource_attr_id == existing_ra.id).first()

            if old_ra_rs is not None and new_ra_rs is None:
                old_ra_rs.resource_attr_id = existing_ra.id

            #mark the RA as deleted so it can be ignored in the remapping
            #process later
            deleted_ras.append(ra_to_remap.id)
            db.DBSession.delete(ra_to_remap)


    for ra_to_remap in ras_to_remap:
        #no need to remap as it's been deleted
        if ra_to_remap.id in deleted_ras:
            continue
        ra_to_remap.attr_id = new_attr_id

    tas_to_remap = db.DBSession.query(TypeAttr)\
        .filter(TypeAttr.attr_id == old_attr_id).all()

    for ta_to_remap in tas_to_remap:
        #is there an existing type attr on the type with the same attr?
        existing_ta = db.DBSession.query(TypeAttr)\
            .filter(TypeAttr.attr_id == new_attr_id,
                    TypeAttr.type_id == ta_to_remap.type_id).first()

        #if so, delete it
        if existing_ta is not None:
            db.DBSession.delete(ta_to_remap)
        else:
            ta_to_remap.attr_id = new_attr_id

    if flush is True:
        db.DBSession.flush()



import sys

import pytest


def test_safe_on_non_windows_platforms():
    """
      Verifies that a Warning is raised on non-Windows platforms
      when attempting to import a Windows-only util and that
      import succeeds on Windows
    """
    if sys.platform.startswith("win"):
        from hydra_base.util.windows import win_get_common_documents
    else:
        with pytest.raises(Warning):
            from hydra_base.util.windows import win_get_common_documents


@pytest.mark.skipif(not sys.platform.startswith("win"), reason="Skipping Windows platform tests")
class TestWindows:

    def test_win_get_common_documents(self):
        """
          Do the Windows utils import correctly, find an appropriate
          windll function, and successfully call this to identify
          a path?
        """
        from hydra_base.util.windows import win_get_common_documents
        common_doc_path = win_get_common_documents()
        assert isinstance(str, common_doc_path)
        assert len(common_doc_path) > 0




"""node_group_link_id

Revision ID: f894faeb055d
Revises: 74ab69a8dcb6
Create Date: 2018-03-29 11:35:42.103923

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql
import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = 'f894faeb055d'
down_revision = '74ab69a8dcb6'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tNode

        try:
            op.alter_column('tNode', 'node_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tNode', 'node_name', new_column_name='name', existing_type=sa.String(200))
            op.alter_column('tNode', 'node_description', new_column_name='description', existing_type=sa.String(1000))
            op.alter_column('tNode', 'node_x', new_column_name='x', existing_type=sa.Float(precision=10, asdecimal=True))
            op.alter_column('tNode', 'node_y', new_column_name='y', existing_type=sa.Float(precision=10, asdecimal=True))
        except Exception as e:
            log.exception(e)

        # ### tLink
        try:

            op.alter_column('tLink', 'link_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tLink', 'link_name', new_column_name='name', existing_type=sa.String(200))
            op.alter_column('tLink', 'link_description', new_column_name='description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)

        # ### tResourceGroup
        try:

            op.alter_column('tResourceGroup', 'group_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tResourceGroup', 'group_name', new_column_name='name', existing_type=sa.String(200))
            op.alter_column('tResourceGroup', 'group_description', new_column_name='description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tNode
        try:
            op.drop_table('tNode_new')
        except:
            log.info("tNode_new isn't there")

        try:
            op.create_table(
                'tNode_new',
                sa.Column('id', sa.Integer, primary_key=True),
                sa.Column('name', sa.Text(200), nullable=False),
                sa.Column('description', sa.Text(1000)),
                sa.Column('layout', sa.Text(1000)),
                sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'),  nullable=False),
                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('x', sa.Float(precision=10, asdecimal=True)),
                sa.Column('y', sa.Float(precision=10, asdecimal=True)),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.UniqueConstraint('name', 'network_id', name="unique node name")
            )

            op.execute("insert into tNode_new (id, name, description, layout, network_id, status, cr_date, x, y) select node_id, node_name, node_description, layout, network_id, status, cr_date, node_x, node_y from tNode")

            op.rename_table('tNode','tNode_old')
            op.rename_table('tNode_new', 'tNode')
            op.drop_table('tNode_old')

        except Exception as e:
            log.exception(e)

        # ## tLink
        try:
            op.drop_table('tLink_new')
        except:
            log.info("tLink_new isn't there")

        try:

            op.create_table(
                'tLink_new',
                sa.Column('id', sa.Integer, primary_key=True),
                sa.Column('name', sa.Text(200), nullable=False),
                sa.Column('description', sa.Text(1000)),
                sa.Column('layout', sa.Text(1000)),
                sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'),  nullable=False),
                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('node_1_id', sa.Integer(), sa.ForeignKey('tNode.id'), nullable=False),
                sa.Column('node_2_id', sa.Integer(), sa.ForeignKey('tNode.id'), nullable=False),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.UniqueConstraint('name', 'network_id', name="unique link name")
            )

            op.execute("insert into tLink_new (id, name, description, layout, network_id, status, cr_date, node_1_id, node_2_id) select link_id, link_name, link_description, layout, network_id, status, cr_date, node_1_id, node_2_id from tLink")

            op.rename_table('tLink','tLink_old')
            op.rename_table('tLink_new', 'tLink')
            op.drop_table('tLink_old')

        except Exception as e:
            log.exception(e)

        # ## tResourceGroup
        try:
            op.drop_table('tResourceGroup_new')
        except:
            log.info("tResourceGroup_new isn't there")

        try:

            op.create_table(
                'tResourceGroup_new',
                sa.Column('id', sa.Integer, primary_key=True),
                sa.Column('name', sa.Text(200), nullable=False),
                sa.Column('description', sa.Text(1000)),
                sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'),  nullable=False),
                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.UniqueConstraint('name', 'network_id', name="unique resourcegroup name")
            )

            op.execute("insert into tResourceGroup_new (id, name, description, network_id, status, cr_date) select group_id, group_name, group_description, network_id, status, cr_date, from tResourceGroup")

            op.rename_table('tResourceGroup','tResourceGroup_old')
            op.rename_table('tResourceGroup_new', 'tResourceGroup')
            op.drop_table('tResourceGroup_old')

        except Exception as e:
            log.exception(e)


def downgrade():

    if op.get_bind().dialect.name == 'mysql':

        # ### tNode

        op.alter_column('tNode', 'id', new_column_name='node_id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
        op.alter_column('tNode', 'name', new_column_name='node_name', existing_type=sa.String(200))

        op.alter_column('tNode', 'description', new_column_name='node_description', existing_type=sa.String(1000))
        op.alter_column('tNode', 'x', new_column_name='node_x', existing_type=sa.String(1000))
        op.alter_column('tNode', 'y', new_column_name='node_y', existing_type=sa.String(1000))

        # ### tLink

        op.alter_column('tLink', 'id', new_column_name='link_id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
        op.alter_column('tLink', 'name', new_column_name='link_name', existing_type=sa.String(200))
        op.alter_column('tLink', 'description', new_column_name='link_description', existing_type=sa.String(1000))

        # ### tResourceGroup

        op.alter_column('tResourceGroup', 'id', new_column_name='group_id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
        op.alter_column('tResourceGroup', 'name', new_column_name='group_name', existing_type=sa.String(200))
        op.alter_column('tResourceGroup', 'description', new_column_name='group_description', existing_type=sa.String(1000))

    else:
        # ## tNode
        op.create_table(
            'tNode_new',
            sa.Column('node_id', sa.Integer, primary_key=True),
            sa.Column('node_name', sa.Text(200), nullable=False),
            sa.Column('node_description', sa.Text(1000)),
            sa.Column('layout', sa.Text(1000)),
            sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'),  nullable=False),
            sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
            sa.Column('x', sa.Float(precision=10, asdecimal=True)),
            sa.Column('y', sa.Float(precision=10, asdecimal=True)),
            sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
            sa.UniqueConstraint('node_name', 'network_id', name="unique node name")
        )

        op.execute("insert into tNode_new (node_id, node_name, node_description, layout, network_id, status, cr_date, node_x, node_y) select id, name, description, layout, network_id, status, cr_date, x, y from tNode")

        op.rename_table('tNode','tNode_old')
        op.rename_table('tNode_new', 'tNode')

        # ## tLink
        op.create_table(
            'tLink_new',
            sa.Column('link_id', sa.Integer, primary_key=True),
            sa.Column('link_name', sa.Text(200), nullable=False),
            sa.Column('link_description', sa.Text(1000)),
            sa.Column('layout', sa.Text(1000)),
            sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'),  nullable=False),
            sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
            sa.Column('node_1_id', sa.Integer(), sa.ForeignKey('tNode.id'), nullable=False),
            sa.Column('node_2_id', sa.Integer(), sa.ForeignKey('tNode.id'), nullable=False),
            sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
            sa.UniqueConstraint('link_name', 'network_id', name="unique link name")
        )

        op.execute("insert into tLink_new (link_id, link_name, link_description, layout, network_id, status, cr_date, node_1_id, node_2_id) select id, name, description, layout, network_id, status, cr_date, node_1_id, node_2_id from tLink")

        op.rename_table('tLink','tLink_old')
        op.rename_table('tLink_new', 'tLink')


        # ## tResourceGroup
        op.create_table(
            'tResourceGroup_new',
            sa.Column('group_id', sa.Integer, primary_key=True),
            sa.Column('group_name', sa.Text(200), nullable=False),
            sa.Column('group_description', sa.Text(1000)),
            sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'),  nullable=False),
            sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
            sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
            sa.UniqueConstraint('group_name', 'network_id', name="unique resourcegroup name")
        )

        op.execute("insert into tResourceGroup_new (group_id, group_name, group_description, network_id, status, cr_date) select id, name, description, network_id, status, cr_date, from tResourceGroup")

        op.rename_table('tResourceGroup','tResourceGroup_old')
        op.rename_table('tResourceGroup_new', 'tResourceGroup')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

time_map = {
    'picosecond'  : 'ps',
    'picoseconds' : 'ps',
    'picosec'     : 'ps',
    'picosecs'    : 'ps',
    'ps'          : 'ps',

    'nanosecond'  : 'ns',
    'nanoseconds' : 'ns',
    'nanosecs'    : 'ns',
    'nanosec'     : 'ns',
    'ns'          : 'ns',

    'microsecond'  : 'µs',
    'microseconds' : 'µs',
    'microsec'     : 'µs',
    'microsecs'    : 'µs',
    'µs'           : 'µs',

    'millisecond'   : 'ms',
    'milliseconds'  : 'ms',
    'millisec'      : 'ms',
    'millisecs'     : 'ms',
    'ms'            : 'ms',

    'second'  : 's',
    'seconds' : 's',
    'sec'     : 's',
    'secs'    : 's',
    's'       : 's',
    
    'minute'  : 'min',
    'minutes' : 'min',
    'min'     : 'min',
    'mins'    : 'min',
    'm'       : 'min',

    'h'     : 'h',
    'hour'  : 'h',
    'hours' : 'h',

    'd'    : 'day',
    'day'  : 'day',
    'days' : 'day',

    'mon'    : 'mon',
    'month'  : 'mon',
    'months' : 'mon',
    'mons'   : 'mon',

    'y'    : 'yr',
    'yr'   : 'yr',
    'year' : 'yr',
    'years': 'yr',
    'yrs'  : 'yr',
}





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from .base import *


__all__ = ['Unit', 'Dimension']

class Unit(Base, Inspect):
    """
    """

    __tablename__='tUnit'

    __table_args__ = (
        UniqueConstraint('abbreviation', name="unique abbreviation"),
    )

    id = Column(Integer(), primary_key=True, nullable=False)
    dimension_id = Column(Integer(), ForeignKey('tDimension.id'), nullable=False)

    # These lines are commented because sqllite seem not accepting utf8_bin. Find a solution
    #     name = Column(Unicode(60, collation='utf8_bin'),  nullable=False)
    #     abbreviation = Column(Unicode(60, collation='utf8_bin'),  nullable=False)
    #     lf = Column(Unicode(60, collation='utf8_bin'),  nullable=True)
    #     cf = Column(Unicode(60, collation='utf8_bin'),  nullable=True)
    #     description = Column(Unicode(1000, collation='utf8_bin'))
    name = Column(Unicode(60),  nullable=False)
    abbreviation = Column(Unicode(60).with_variant(mysql.VARCHAR(60, collation='utf8_bin'), 'mysql'),  nullable=False)
    lf = Column(Unicode(60),  nullable=True)
    cf = Column(Unicode(60),  nullable=True)
    description = Column(Unicode(1000))

    project_id = Column(Integer(), ForeignKey('tProject.id'), index=True, nullable=True)

    dimension = relationship('Dimension', backref=backref("units", uselist=True, order_by=dimension_id, cascade="all, delete-orphan"), lazy='joined')
    project   = relationship('Project', backref=backref("units", order_by=dimension_id, cascade="all, delete-orphan"), lazy='joined')

    _parents  = ['tDimension', 'tProject']
    _children = ['tDataset', 'tTypeAttr']

    def __repr__(self):
        return "{0}".format(self.abbreviation)


class Dimension(Base, Inspect):
    """
    """

    __tablename__='tDimension'

    id = Column(Integer(), primary_key=True, nullable=False)

    # These lines are commented because sqllite seem not accepting utf8_bin. Find a solution
    # name = Column(Unicode(60, collation='utf8_bin'),  nullable=False, unique=True)
    # description = Column(Unicode(1000, collation='utf8_bin'))

    name = Column(Unicode(60),  nullable=False, unique=True)
    description = Column(Unicode(1000))

    project_id = Column(Integer(), ForeignKey('tProject.id'), index=True, nullable=True)

    _parents  = ['tProject']
    _children = ['tUnit', 'tAttr']

    def __repr__(self):
        return "{0}".format(self.name)

def create_resourcedata_view():
    from .network import ResourceAttr, ResourceScenario
    from .dataset import Dataset
    from .attributes import Attr
    #These are for creating the resource data view (see bottom of page)
    from sqlalchemy import select
    from sqlalchemy.schema import DDLElement
    from sqlalchemy.sql import table
    from sqlalchemy.ext import compiler

    class CreateView(DDLElement):
        def __init__(self, name, selectable):
            self.name = name
            self.selectable = selectable

    class DropView(DDLElement):
        def __init__(self, name):
            self.name = name

    @compiler.compiles(CreateView)
    def compile(element, compiler, **kw):
        return "CREATE VIEW %s AS %s" % (element.name, compiler.sql_compiler.process(element.selectable))

    @compiler.compiles(DropView)
    def compile(element, compiler, **kw):
        return "DROP VIEW %s" % (element.name)

    def view(name, metadata, selectable):
        t = table(name)

        for c in selectable.c:
            c._make_proxy(t)

        CreateView(name, selectable).execute_at('after-create', metadata)
        DropView(name).execute_at('before-drop', metadata)
        return t


    view_qry = select([
        ResourceAttr.id,
        ResourceAttr.attr_id,
        Attr.name,
        ResourceAttr.id,
        ResourceAttr.network_id,
        ResourceAttr.node_id,
        ResourceAttr.link_id,
        ResourceAttr.group_id,
        ResourceScenario.scenario_id,
        ResourceScenario.dataset_id,
        Dataset.unit_id,
        Dataset.name,
        Dataset.type,
        Dataset.value]).where(ResourceScenario.resource_attr_id==ResourceAttr.attr_id).where(ResourceAttr.attr_id==Attr.id).where(ResourceScenario.dataset_id==Dataset.id)

    stuff_view = view("vResourceData", Base.metadata, view_qry)
#TODO: Understand why this view is not being created.
#create_resourcedata_view()



"""lengthen names

Revision ID: 52d59176f0da
Revises: 6883f6e39176
Create Date: 2018-12-27 16:07:41.052385

"""
from alembic import op
import sqlalchemy as sa

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '52d59176f0da'
down_revision = '6883f6e39176'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':


        try:
            op.alter_column('tTemplate', 'name', existing_type=sa.String(60), type_=sa.String(200), nullable=False, unique=True)
            op.add_column('tTemplate', sa.Column('description', sa.String(1000)))
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tDataset_new')
        except:
            log.info("tDataset_new isn't there")

        try:
            op.create_table(
                'tTemplate_new',
                sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
                sa.Column('name', sa.String(200),  nullable=False, unique=True),
                sa.Column('description', sa.String(1000)),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('layout', sa.Text().with_variant(sa.mysql.TEXT(4294967295), 'mysql'),  nullable=True)
            )

            op.execute("insert into tTemplate_new (id, name, description, cr_date, layout) select id, name, '', cr_date, layout from tTemplate")

            op.rename_table('tTemplate','tTemplate_old')
            op.rename_table('tTemplate_new', 'tTemplate')
            op.drop_table('tTemplate_old')

        except Exception as e:
            log.exception(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':
        try:
            op.alter_column('tTemplate', 'name', existing_type=sa.String(60), type_=sa.String(200), nullable=False, unique=True)
            op.drop_column('tTemplate',  'description')
        except Exception as e:
            log.exception(e)

    else: ## sqlite
        try:
            op.create_table(
                'tTemplate_new',
                sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
                sa.Column('name', sa.String(60),  nullable=False, unique=True),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('layout', sa.Text().with_variant(sa.mysql.TEXT(4294967295), 'mysql'),  nullable=True)
            )

            op.execute("insert into tTemplate_new (id, name, cr_date, layout) select id, name, cr_date, layout from tTemplate")

            op.rename_table('tTemplate','tTemplate_old')
            op.rename_table('tTemplate_new', 'tTemplate')
            op.drop_table('tTemplate_old')

        except Exception as e:
            log.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
import datetime
import copy
import pytest

import hydra_base as hb
from hydra_base.lib.objects import JSONObject

import logging
log = logging.getLogger(__name__)

class TestProject:
    """
        Test for working with projects in Hydra
    """

    def add_attributes(self, client, proj):
        #Create some attributes, which we can then use to put data on our nodes
        attr1 = client.testutils.create_attribute("proj_attr_1")
        attr2 = client.testutils.create_attribute("proj_attr_2")
        attr3 = client.testutils.create_attribute("proj_attr_3")

        proj_attr_1  = JSONObject({})
        proj_attr_1.id = -1
        proj_attr_1.attr_id = attr1.id
        proj_attr_2  = JSONObject({})
        proj_attr_2.attr_id = attr2.id
        proj_attr_2.id = -2
        proj_attr_3  = JSONObject({})
        proj_attr_3.attr_id = attr3.id
        proj_attr_3.id = -3

        attributes = []

        attributes.append(proj_attr_1)
        attributes.append(proj_attr_2)
        attributes.append(proj_attr_3)

        proj.attributes = attributes

        return proj

    def add_data(self, client, project):

        attribute_data = []

        attrs = project.attributes

        attribute_data.append(client.testutils.create_descriptor(attrs[0], val="just project desscriptor"))
        attribute_data.append(client.testutils.create_array(attrs[1]))
        attribute_data.append(client.testutils.create_timeseries(attrs[2]))

        project.attribute_data = attribute_data

        return project

    def test_update(self, client, network_with_data):
        """
            The network here is necessary for a scenario with ID 1 to exist.
            Under normal circumstances, scenario 1 will always exist, as it's an initial requirement
            of the database setup.
        """
        project = JSONObject({})
        project.name = 'SOAP test %s'%(datetime.datetime.now())
        project.description = \
            'A project created through the SOAP interface.'
        project.appdata = {'test':'metadata'}
        project = self.add_attributes(client, project)
        project = self.add_data(client, project)

        new_project_i = client.add_project(project)

        project_j = client.get_project(new_project_i.id)

        assert project_j.appdata['test'] == 'metadata'

        new_project = copy.deepcopy(project_j)

        new_project.description = \
            'An updated project created through the Hydra Base interface.'

        new_project.appdata['test1'] = 'metadata1'

        updated_project = client.update_project(new_project)

        assert updated_project.appdata['test'] == 'metadata'
        assert updated_project.appdata['test1'] == 'metadata1'

        assert project_j.id == updated_project.id, \
            "project_id changed on update."
        assert project_j.created_by is not None, \
            "created by is null."
        assert project_j.name == updated_project.name, \
            "project_name changed on update."
        assert project_j.description != updated_project.description,\
            "project_description did not update"
        assert updated_project.description == \
            'An updated project created through the Hydra Base interface.', \
            "Update did not work correctly."

        rs_to_check = client.get_project_attribute_data(new_project.id)[0]
        assert rs_to_check.dataset.type == 'descriptor' and \
               rs_to_check.dataset.value == 'just project desscriptor', \
               "There is an inconsistency with the attributes."

    def test_load(self, client):
        project = JSONObject({})
        project.name = 'Test Project %s'%(datetime.datetime.now())
        project.description = \
            'A project created through the SOAP interface.'
        project = client.add_project(project)

        new_project = client.get_project(project.id)

        assert new_project.name == project.name, \
            "project_name is not loaded correctly."
        assert project.description == new_project.description,\
            "project_description did not load correctly."

    def test_set_project_status(self, client):
        project = JSONObject({})
        project.name = 'SOAP test %s'%(datetime.datetime.now())
        project.description = \
            'A project created through the SOAP interface.'
        project = client.add_project(project)

        client.set_project_status(project.id, 'X')

        proj = client.get_project(project.id)

        assert proj.status == 'X', \
            'Deleting project did not work correctly.'

    def test_delete(self, client, network_with_data):
        net = network_with_data
        project_id = net.project_id
        log.info("Purging project %s", project_id)
        res = client.delete_project(project_id)

        assert res == 'OK'
        log.info("Trying to get project %s. Should fail.",project_id)
        with pytest.raises(hb.HydraError):
            client.get_project(project_id)

    def test_get_projects(self, client):

        project = JSONObject({})

        project.name = 'SOAP test %s'%(datetime.datetime.now())
        project.description = \
            'A project created through the SOAP interface.'

        project = client.add_project(project)

        projects = client.get_projects(pytest.root_user_id)

        assert len(projects) > 0, "Projects for user were not retrieved."

        assert projects[0].status == 'A'

    def test_get_networks(self, client, projectmaker, networkmaker):

        proj = projectmaker.create()

        net1 = networkmaker.create(project_id=proj.id)

        net2 = networkmaker.create(project_id=proj.id)
        nets = client.get_networks(proj.id)

        test_net = nets[0]
        assert test_net.scenarios is not None
        test_scenario = test_net.scenarios[0]
        assert len(test_net.nodes) > 0
        assert len(test_net.links) > 0
        assert len(test_scenario.resourcescenarios) == 0


        assert len(nets) == 2, "Networks were not retrieved correctly"

        nets = client.get_networks(proj.id, include_data=True)

        test_scenario = nets[0].scenarios[0]
        assert len(test_scenario.resourcescenarios) > 0

    def test_get_all_project_owners(self, client, projectmaker):
        proj = projectmaker.create()

        projectowners = client.get_all_project_owners()

        #there should be at LEAST 4, owing to the project for this test
        assert len(projectowners) >= 4

        projectowners = client.get_all_project_owners([proj.id])

        assert len(projectowners) == 4

        with pytest.raises(hb.exceptions.HydraError):
            #check for non-admin users
            client.user_id = 5
            projectowners = client.get_all_project_owners([proj.id])
        #set back to admin
        client.user_id=1

    def test_bulk_set_project_owners(self, client, projectmaker):
        proj = projectmaker.create(share=False)

        projectowners = client.get_all_project_owners([proj.id])

        assert len(projectowners) == 1

        projectowners = client.get_all_project_owners([proj.id])

        new_owner = JSONObject(dict(
            project_id=proj.id,
            user_id=2,
            view='Y',
            edit='Y',
            share='Y',
        ))

        client.bulk_set_project_owners([new_owner])

        projectowners = client.get_all_project_owners([proj.id])

        assert len(projectowners) == 2

    def test_rename_status_x_project(self, client, projectmaker, networkmaker):
        sender_user_id = client.user_id
        proj = projectmaker.create()
        net1 = networkmaker.create(project_id=proj.id)
        net2 = networkmaker.create(project_id=proj.id)

        proj_name = proj.name

        #Set the status of the project to 'X'.
        client.set_project_status(proj.id, 'X')
        #now create a project with the name of the deleted project
        project = JSONObject({})
        project.name = proj_name
        project.description = \
            'A project created with the same name as a deleted project.'
        project = client.add_project(project)

        old_proj = client.get_project(proj.id)

        assert old_proj.status == 'X'
        assert old_proj.name != project.name



    def test_clone_project(self, client, projectmaker, networkmaker):

        proj = projectmaker.create()

        net1 = networkmaker.create(project_id=proj.id)

        net2 = networkmaker.create(project_id=proj.id)

        recipient_user = client.get_user_by_name('UserA')

        new_project_name = 'New Project'

        cloned_project_id = client.clone_project(
            proj.id,
            recipient_user_id=recipient_user.id,
            new_project_name=new_project_name)

        cloned_project = client.get_project(cloned_project_id)

        assert cloned_project.name == new_project_name
        cloned_networks = client.get_networks(cloned_project.id)
        assert len(cloned_networks) == 2

        #check with no name provided
        cloned_project_id = client.clone_project(
            proj.id,
            recipient_user_id=recipient_user.id,
            new_project_name=None)

        cloned_project = client.get_project(cloned_project_id)

        assert cloned_project.name.find('Cloned') > 0
        cloned_networks = client.get_networks(cloned_project.id)
        assert len(cloned_networks) == 2

    def test_clone_project(self, client, projectmaker, networkmaker):

        proj = projectmaker.create()

        net1 = networkmaker.create(project_id=proj.id)

        net2 = networkmaker.create(project_id=proj.id)

        recipient_user = client.get_user_by_name('UserA')

        new_project_name = 'New Project'

        cloned_project_id = client.clone_project(
            proj.id,
            recipient_user_id=recipient_user.id,
            new_project_name=new_project_name)

        cloned_project = client.get_project(cloned_project_id)

        assert cloned_project.name == new_project_name
        cloned_networks = client.get_networks(cloned_project.id)
        assert len(cloned_networks) == 2

        #check with no name provided
        cloned_project_id = client.clone_project(
            proj.id,
            recipient_user_id=recipient_user.id,
            new_project_name=None)

        cloned_project = client.get_project(cloned_project_id)

        assert cloned_project.name.find('Cloned') > 0
        cloned_networks = client.get_networks(cloned_project.id)
        assert len(cloned_networks) == 2

    def test_clone_project_to_other_user(self, client, projectmaker, networkmaker):
        sender_user_id = client.user_id
        proj = projectmaker.create()
        net1 = networkmaker.create(project_id=proj.id)
        net2 = networkmaker.create(project_id=proj.id)

        #"sharing with a non-admin user"
        recipient_user = client.get_user_by_id(pytest.user_c.id)

        sender_projects_before = client.get_projects(client.user_id)

        client.user_id = recipient_user.id
        recipient_projects_before = client.get_projects(client.user_id)

        client.user_id = sender_user_id

        new_project_name = 'Cloned Project'

        test_delete_project_id = client.clone_project(
            proj.id,
            recipient_user_id=recipient_user.id,
            new_project_name=new_project_name)

        #Check the recipient has received the project
        client.user_id = recipient_user.id
        recipient_user_project = client.get_project(test_delete_project_id)
        assert len(recipient_user_project.networks) == 2

        recipient_projects_after = client.get_projects(client.user_id)

        assert len(recipient_projects_after) == len(recipient_projects_before)+1

        client.user_id = sender_user_id
        sender_projects_after = client.get_projects(client.user_id)
        assert len(sender_projects_after) == len(sender_projects_before)

        #check with no name provided
        cloned_project_id = client.clone_project(
            proj.id,
            recipient_user_id=recipient_user.id,
            new_project_name=None)

        cloned_project = client.get_project(cloned_project_id)

        assert cloned_project.name.find('Cloned') > 0
        cloned_networks = client.get_networks(cloned_project.id)
        assert len(cloned_networks) == 2

        #Check that the resource attribute references have been updated correctly
        #by comparing the attr_id on the original network and the cloned network for the same attribute name,
        #ensuring they are not the same.
        for n in cloned_project.networks:
            net = client.get_network(network_id=n.id)
            for ra in net.attributes:
                a = client.get_attribute_by_id(attr_id=ra.attr_id)
                if a.network_id is not None:
                    for ra1 in net1.attributes:
                        if ra1.name == a.name:
                            assert ra1.network_id is not None
                            assert ra.attr_id != ra1.attr_id

    def test_rename_of_status_x_projects_in_clone(self, client, projectmaker, networkmaker):
        sender_user_id = client.user_id
        proj = projectmaker.create()
        net1 = networkmaker.create(project_id=proj.id)
        net2 = networkmaker.create(project_id=proj.id)

        #"sharing with a non-admin user"
        recipient_user = client.get_user_by_id(pytest.user_c.id)

        sender_projects_before = client.get_projects(client.user_id)

        client.user_id = recipient_user.id
        recipient_projects_before = client.get_projects(client.user_id)

        client.user_id = sender_user_id

        new_project_name = 'Cloned Project Test'

        test_delete_project_id = client.clone_project(
            proj.id,
            recipient_user_id=recipient_user.id,
            new_project_name=new_project_name)

        #test renaming of deleted projects
        client.set_project_status(test_delete_project_id, 'X')
        changed_project = client.get_project(test_delete_project_id)
        assert changed_project.status == 'X'

        cloned_project_id = client.clone_project(
            proj.id,
            recipient_user_id=recipient_user.id,
            new_project_name=new_project_name)

        renamed_project = client.get_project(test_delete_project_id)
        assert renamed_project.status == 'X'

    def test_get_project_by_network_id(self, client, projectmaker, networkmaker):

        proj = projectmaker.create()

        net1 = networkmaker.create(project_id=proj.id)

        net2 = networkmaker.create(project_id=proj.id)
        nets = client.get_networks(proj.id)


        project_i = client.get_project_by_network_id(net1.id)

        assert project_i.id == proj.id

    def test_share_project(self, client, projectmaker):

        proj_user = client.user_id
        proj = projectmaker.create(share=False)
        client.user_id = pytest.user_c.id
        with pytest.raises(hb.HydraError):
            client.get_project(proj.id)

        client.user_id = proj_user
        client.share_project(proj.id, ['UserC'], False, False)

        #Get the project as user C -- this should succeeed
        client.user_id = pytest.user_c.id
        client.get_project(proj.id)

        #now revolk access
        client.user_id = proj_user
        client.unshare_project(proj.id, ['UserC'])

        #user c no longer has access
        client.user_id = pytest.user_c.id
        with pytest.raises(hb.HydraError):
            client.get_project(proj.id)

    def test_set_project_status_cache_invalidation(self, client, projectmaker):
        proj = projectmaker.create()
        project_id = proj.id

        proj1 = client.get_project(project_id)
        assert proj1.status == 'A'

        client.set_project_status(project_id, 'X')

        proj2 = client.get_project(project_id)
        assert proj2.status == 'X'

        #get projecs again to check the project is no longer present
        projects = client.get_projects(pytest.root_user_id)
        assert project_id not in [p.id for p in projects]


"""project-inheritance

Revision ID: 04e4ae80b7b9
Revises: cec2b77ad85e
Create Date: 2022-01-10 10:00:02.614468

"""
from alembic import op
import sqlalchemy as sa

import logging

log = logging.getLogger(__name__)


# revision identifiers, used by Alembic.
revision = '04e4ae80b7b9'
down_revision = 'a81a860cda39'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.add_column('tProject', sa.Column('parent_id', sa.Integer(), sa.ForeignKey('tProject.id'), nullable=True))
        except Exception as e:
            log.critical(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':
        try:
            op.drop_column('tProject', 'parent_id')
        except Exception as e:
            log.exception(e)



import json
import pytest
import random

from packaging import version

import hydra_base
from hydra_base.lib.storage import MongoStorageAdapter
from hydra_base.lib.objects import Dataset

# Minimum required mongo version
mongo_min_version = version.parse("3.6.8")


@pytest.fixture
def mongo():
    mongo = MongoStorageAdapter()
    return mongo

@pytest.fixture
def mongo_config():
    """ This can be parametrized to allow different config sources to be tested """
    return MongoStorageAdapter.get_mongo_config()


@pytest.mark.externaldb
class TestMongo():
    def generate_datasets(self, client, mongo_config, num_datasets=12, large_step=3):
        """
        Returns `num_datasets` Datasets containing an array of random floats.
        Every `large_step`'th dataset (including the 0'th) is larger
        than the external storage threshold.
        """
        mongo_threshold = mongo_config["threshold"]
        datasets = []
        unit_id = client.get_unit_by_abbreviation("m s^-1").id
        for idx in range(num_datasets):
            ds = Dataset()
            ds.name = f"Dataset {idx}"
            ds.type = "ARRAY"
            ds.unit_id = unit_id
            if idx % large_step:
                data_sz = 10
            else:
                data_sz = mongo_threshold * random.randint(1, 3) + 1
            ds.value = [random.uniform(1, 100) for _ in range(data_sz)]
            datasets.append(ds)

        return datasets


    def test_server_status(self, mongo):
        """
        Is the server available and can return the getBuildInfo() dict?
        """
        status = mongo.client.server_info()
        assert isinstance(status, dict)


    def test_mongo_version(self, mongo):
        """
        Does the server's version meet the minimum requirement?
        """
        status = mongo.client.server_info()
        semver = status.get("version")
        assert semver, "No version key in build info"
        vp = version.parse(semver)
        assert vp >= mongo_min_version


    def test_mongo_config(self, mongo):
        required = ("threshold", "value_location_key", "direct_location_token")
        mongo_config = MongoStorageAdapter.get_mongo_config()

        for key in required:
            assert key in mongo_config, f"Mongo config missing `{key}` definition"


    def test_bulk_add_mongo_data(self, client, mongo_config, mongo):
        """
        Builds a non-trivial number of large datasets and adds *the values*
        of these directly to mongo using the adapter's `bulk_insert_values`
        func.
        Bulk insertion of datasets via Hydra is tested in test_scenario::bulk_add_data
        """

        mongo_threshold = mongo_config["threshold"]
        num_datasets = 32
        datasets = []
        unit_id = client.get_unit_by_abbreviation("m s^-1").id
        for idx in range(num_datasets):
            ds = Dataset()
            ds.name = f"Bulk dataset {idx}"
            ds.type = "ARRAY"
            ds.unit_id = unit_id
            data_sz = mongo_threshold * random.randint(1, 3) + 1
            ds.value = [random.uniform(1, 100) for _ in range(data_sz)]
            datasets.append(ds)

        inserted = mongo.bulk_insert_values([d.value for d in datasets])
        assert len(inserted.inserted_ids) == num_datasets, "Datasets insertion count mismatch"
        """ Remove data from test collection """
        for _id in inserted.inserted_ids:
            mongo.delete_document_by_object_id(_id)


    def test_add_datasets(self, client, mongo_config, mongo):
        """
        Are datasets with size exceeding `mongo_threshold` correctly routed to
        external storage with appropriate metadata added to indicate this?
        """
        mongo_threshold = mongo_config["threshold"]
        mongo_location_key = mongo_config["value_location_key"]
        mongo_location_external = mongo_config["direct_location_token"]

        def lookup_dataset_metadata(ds_id):
            for m in metadatas:
                if m["dataset_id"] == ds_id:
                    return m

        datasets = self.generate_datasets(client, mongo_config)
        added = []
        for ds in datasets:
            added.append(client.add_dataset(ds.type, ds.value, ds.unit_id, {}, ds.name, flush=True))

        metadatas = client.get_metadata([ds.id for ds in added])

        """ Do external datasets and only these have location metadata? """
        for ds in added:
            if len(ds.value) > mongo_threshold:
                m = lookup_dataset_metadata(ds.id)
                key = m.get("key")
                assert key == mongo_location_key, "Location key missing from large dataset"
                location = m.get("value")
                assert location == mongo_location_external, "Invalid location metadata value"
            else:
                m = lookup_dataset_metadata(ds.id)
                assert not m

        """ Are the added values transparently retrieved for every dataset? """
        for local_ds, added_ds in zip(datasets, added):
            assert local_ds.value == added_ds.value, "Dataset value mismatch"


    def test_shrink_dataset(self, client, mongo_config, mongo):
        """
        Is a large dataset initially placed in external storage, then relocated
        to the SQL db if its size is reduced beneath the external threshold?
        """
        mongo_location_key = mongo_config["value_location_key"]
        mongo_location_external = mongo_config["direct_location_token"]
        mongo_threshold = mongo_config["threshold"]
        unit_id = client.get_unit_by_abbreviation("m s^-1").id

        ds = Dataset()
        ds.name = "Shrinking dataset"
        ds.type = "ARRAY"
        ds.unit_id = unit_id
        data_sz = mongo_threshold * random.randint(1, 3) + 1
        ds.value = [random.uniform(1, 100) for _ in range(data_sz)]

        server_ds = client.add_dataset(ds.type, ds.value, ds.unit_id, {}, ds.name, flush=True)
        metadata = client.get_metadata([server_ds.id])

        key = metadata[0].get("key")
        assert key == mongo_location_key, "Location key missing from large dataset"
        location = metadata[0].get("value")
        assert location == mongo_location_external, "Invalid location metadata value"

        """ Shrink the dataset and confirm it has been relocated """
        server_ds.value = [random.uniform(1, 100) for _ in range(10)]
        shrunk_ds = client.update_dataset(server_ds.id, server_ds.name, server_ds.type, json.dumps(server_ds.value), server_ds.unit_id, {})
        shrunk_metadata = client.get_metadata([shrunk_ds.id])
        assert not shrunk_metadata


    def test_grow_dataset(self, client, mongo_config, mongo):
        """
        Is a small dataset initially placed in SQL db storage, and then relocated
        to external storage when its size grows to exceeds the external threshold?
        """
        mongo_location_key = mongo_config["value_location_key"]
        mongo_location_external = mongo_config["direct_location_token"]
        mongo_threshold = mongo_config["threshold"]
        unit_id = client.get_unit_by_abbreviation("m s^-1").id

        ds = Dataset()
        ds.name = "Grow dataset"
        ds.type = "ARRAY"
        ds.unit_id = unit_id
        ds.value = [random.uniform(1, 100) for _ in range(10)]

        server_ds = client.add_dataset(ds.type, ds.value, ds.unit_id, {}, ds.name, flush=True)
        metadata = client.get_metadata([server_ds.id])
        """ No storage location, therefore SQL db """
        assert not metadata

        """ Grow the dataset and confirm it has been relocated """
        data_sz = mongo_threshold * random.randint(1, 3) + 1
        server_ds.value = [random.uniform(1, 100) for _ in range(data_sz)]
        grown_ds = client.update_dataset(server_ds.id, server_ds.name, server_ds.type, json.dumps(server_ds.value), server_ds.unit_id, {})

        grown_metadata = client.get_metadata([grown_ds.id])
        key = grown_metadata[0].get("key")
        assert key == mongo_location_key, "Location key missing from large dataset"
        location = grown_metadata[0].get("value")
        assert location == mongo_location_external, "Invalid location metadata value"



# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta
import six
from decimal import Decimal, ROUND_HALF_UP
from dateutil.parser import parse
from .. import config
import pandas as pd
import six
import pytz



import logging
log = logging.getLogger(__name__)

#"2013-08-13 15:55:43.468886Z"
FORMAT = "%Y-%m-%d %H:%M:%S.%f"

"""
    A mapping from commonly used time periods to the appropriate hydra-compatible
    time period abbreviation
"""
from .time_map import time_map


def get_time_period(period_name):
    """
        Given a time period name, fetch the hydra-compatible time
        abbreviation.
    """
    time_abbreviation = time_map.get(period_name.lower())

    if time_abbreviation is None:
        raise Exception("Symbol %s not recognised as a time period"%period_name)

    return time_abbreviation

def get_datetime(timestamp):
    """
        Turn a string timestamp into a date time. First tries to use dateutil.
        Failing that it tries to guess the time format and converts it manually
        using stfptime.

        @returns: A timezone unaware timestamp.
    """
    timestamp_is_float = False
    try:
        float(timestamp)
        timestamp_is_float = True
    except (ValueError, TypeError):
        pass

    if timestamp_is_float == True:
        raise ValueError("Timestamp %s is a float"%(timestamp,))

    #First try to use date util. Failing that, continue
    try:
        parsed_dt = parse(timestamp, dayfirst=False)
        return parsed_dt.astimezone(pytz.utc)
    except:
        pass

    if isinstance(timestamp, datetime):
        return timestamp.astimezone(pytz.utc)

    fmt = guess_timefmt(timestamp)
    if fmt is None:
        fmt = FORMAT

    # and proceed as usual
    try:
        ts_time = datetime.strptime(timestamp, fmt)
    except ValueError as e:
        if e.message.split(' ', 1)[0].strip() == 'unconverted':
            utcoffset = e.message.split()[3].strip()
            timestamp = timestamp.replace(utcoffset, '')
            ts_time = datetime.strptime(timestamp, fmt)
            # Apply offset
            tzoffset = timedelta(hours=int(utcoffset[0:3]),
                                          minutes=int(utcoffset[3:5]))
            ts_time -= tzoffset
        else:
            raise e

    return ts_time


def timestamp_to_ordinal(timestamp):
    """Convert a timestamp as defined in the soap interface to the time format
    stored in the database.
    """

    if timestamp is None:
        return None

    ts_time = get_datetime(timestamp)
    # Convert time to Gregorian ordinal (1 = January 1st, year 1)
    ordinal_ts_time = Decimal(ts_time.toordinal())
    total_seconds = (ts_time -
                     datetime(ts_time.year,
                                       ts_time.month,
                                       ts_time.day,
                                       0, 0, 0)).total_seconds()

    fraction = (Decimal(repr(total_seconds)) / Decimal(86400)).quantize(Decimal('.00000000000000000001'),rounding=ROUND_HALF_UP)
    ordinal_ts_time += fraction
    log.debug("%s converted to %s", timestamp, ordinal_ts_time)

    return ordinal_ts_time


def ordinal_to_timestamp(date):
    if date is None:
        return None

    if isinstance(date, six.string_types):
        try:
            date = Decimal(date)
        except:
            log.warning("Unable to convert %s to a timestamp" % (date))
            return date

    day = int(date)
    time = Decimal(str(date - day))
    time_in_secs_ms = (time * Decimal(86400)).quantize(Decimal('.000001'),
                                                       rounding=ROUND_HALF_UP)

    time_in_secs = int(time_in_secs_ms)
    time_in_ms = int((time_in_secs_ms - time_in_secs) * 1000000)

    td = timedelta(seconds=int(time_in_secs), microseconds=time_in_ms)
    d = datetime.fromordinal(day) + td
    log.debug("%s converted to %s", date, d)

    return get_datetime(d)

def date_to_string(date, seasonal=False):
    """Convert a date to a standard string used by Hydra. The resulting string
    looks like this::

        '2013-10-03 00:49:17.568-0400'

    Hydra also accepts seasonal time series (yearly recurring). If the flag
    ``seasonal`` is set to ``True``, this function will generate a string
    recognised by Hydra as seasonal time stamp.
    """

    seasonal_key = config.get('DEFAULT', 'seasonal_key', '9999')
    if seasonal:
        FORMAT = seasonal_key+'-%m-%dT%H:%M:%S.%f'
    else:
        FORMAT = '%Y-%m-%dT%H:%M:%S.%f'
    return date.strftime(FORMAT)


def guess_timefmt(datestr):
    """
    Try to guess the format a date is written in.

    The following formats are supported:

    ================= ============== ===============
    Format            Example        Python format
    ----------------- -------------- ---------------
    ``YYYY-MM-DD``    2002-04-21     %Y-%m-%d
    ``YYYY.MM.DD``    2002.04.21     %Y.%m.%d
    ``YYYY MM DD``    2002 04 21     %Y %m %d
    ``DD-MM-YYYY``    21-04-2002     %d-%m-%Y
    ``DD.MM.YYYY``    21.04.2002     %d.%m.%Y
    ``DD MM YYYY``    21 04 2002     %d %m %Y
    ``DD/MM/YYYY``    21/04/2002     %d/%m/%Y
    ================= ============== ===============

    These formats can also be used for seasonal (yearly recurring) time series.
    The year needs to be replaced by ``9999`` or another configurable year
    representing the seasonal year..

    The following formats are recognised depending on your locale setting.
    There is no guarantee that this will work.

    ================= ============== ===============
    Format            Example        Python format
    ----------------- -------------- ---------------
    ``DD-mmm-YYYY``   21-Apr-2002    %d-%b-%Y
    ``DD.mmm.YYYY``   21.Apr.2002    %d.%b.%Y
    ``DD mmm YYYY``   21 Apr 2002    %d %b %Y
    ``mmm DD YYYY``   Apr 21 2002    %b %d %Y
    ``Mmmmm DD YYYY`` April 21 2002  %B %d %Y
    ================= ============== ===============

    .. note::
        - The time needs to follow this definition without exception:
            `%H:%M:%S.%f`. A complete date and time should therefore look like
            this::

                2002-04-21 15:29:37.522

        - Be aware that in a file with comma separated values you should not
          use a date format that contains commas.
    """

    if isinstance(datestr, float) or isinstance(datestr, int):
        return None

    seasonal_key = str(config.get('DEFAULT', 'seasonal_key', '9999'))

    #replace 'T' with space to handle ISO times.
    if datestr.find('T') > 0:
        dt_delim = 'T'
    else:
        dt_delim = ' '

    delimiters = ['-', '.', ' ', '/']
    formatstrings = [['%Y', '%m', '%d'],
                     ['%d', '%m', '%Y'],
                     ['%d', '%b', '%Y'],
                     ['XXXX', '%m', '%d'],
                     ['%d', '%m', 'XXXX'],
                     ['%d', '%b', 'XXXX'],
                     [seasonal_key, '%m', '%d'],
                     ['%d', '%m', seasonal_key],
                     ['%d', '%b', seasonal_key]]

    timeformats = ['%H:%M:%S.%f', '%H:%M:%S', '%H:%M', '%H:%M:%S.%f000Z', '%H:%M:%S.%fZ']

    # Check if a time is indicated or not
    for timefmt in timeformats:
        try:
            datetime.strptime(datestr.split(dt_delim)[-1].strip(), timefmt)
            usetime = True
            break
        except ValueError:
            usetime = False

    # Check the simple ones:
    for fmt in formatstrings:
        for delim in delimiters:
            datefmt = fmt[0] + delim + fmt[1] + delim + fmt[2]
            if usetime:
                for timefmt in timeformats:
                    complfmt = datefmt + dt_delim + timefmt
                    try:
                        datetime.strptime(datestr, complfmt)
                        return complfmt
                    except ValueError:
                        pass
            else:
                try:
                    datetime.strptime(datestr, datefmt)
                    return datefmt
                except ValueError:
                    pass

    # Check for other formats:
    custom_formats = ['%d/%m/%Y', '%b %d %Y', '%B %d %Y','%d/%m/XXXX', '%d/%m/'+seasonal_key]

    for fmt in custom_formats:
        if usetime:
            for timefmt in timeformats:
                complfmt = fmt + dt_delim + timefmt
                try:
                    datetime.strptime(datestr, complfmt)
                    return complfmt
                except ValueError:
                    pass

        else:
            try:
                datetime.strptime(datestr, fmt)
                return fmt
            except ValueError:
                pass

    return None

def reindex_timeseries(ts_string, new_timestamps):
    """
        get data for timesamp

        :param a JSON string, in pandas-friendly format
        :param a timestamp or list of timestamps (datetimes)
        :returns a pandas data frame, reindexed with the supplied timestamos or None if no data is found
    """
    #If a single timestamp is passed in, turn it into a list
    #Reindexing can't work if it's not a list
    if not isinstance(new_timestamps, list):
        new_timestamps = [new_timestamps]

    #Convert the incoming timestamps to datetimes
    #if they are not datetimes.
    new_timestamps_converted = []
    for t in new_timestamps:
        new_timestamps_converted.append(get_datetime(t))

    new_timestamps = new_timestamps_converted

    seasonal_year = config.get('DEFAULT','seasonal_year', '1678')
    seasonal_key = config.get('DEFAULT', 'seasonal_key', '9999')

    ts = ts_string.replace(seasonal_key, seasonal_year)

    timeseries = pd.read_json(ts)

    idx = timeseries.index

    ts_timestamps = new_timestamps

    #'Fix' the incoming timestamp in case it's a seasonal value
    if type(idx) == pd.DatetimeIndex:
        if set(idx.year) == set([int(seasonal_year)]):
            if isinstance(new_timestamps,  list):
                seasonal_timestamp = []
                for t in ts_timestamps:
                    t_1900 = t.replace(year=int(seasonal_year))
                    seasonal_timestamp.append(t_1900)
                ts_timestamps = seasonal_timestamp

    #Reindex the timeseries to reflect the requested timestamps
    reindexed_ts = timeseries.reindex(ts_timestamps, method='ffill')

    i = reindexed_ts.index

    reindexed_ts.index = pd.Index(new_timestamps, names=i.names)

    #If there are no values at all, just return None
    if len(reindexed_ts.dropna()) == 0:
        return None

    #Replace all numpy NAN values with None
    pandas_ts = reindexed_ts.where(reindexed_ts.notnull(), None)

    return pandas_ts

def parse_time_step(time_step, target='s', units_ref=None):
    """
        Read in the time step and convert it to seconds.
    """
    log.info("Parsing time step %s", time_step)
    # export numerical value from string using regex
    value = re.findall(r'\d+', time_step)[0]
    valuelen = len(value)

    try:
        value = float(value)
    except:
        HydraPluginError("Unable to extract number of time steps (%s) from time step %s" % (value, time_step))

    unit = time_step[valuelen:].strip()

    period = get_time_period(unit)

    log.info("Time period is %s", period)

    converted_time_step = units_ref.convert(value, period, target)

    log.info("Time period is %s %s", converted_time_step, period)

    return float(converted_time_step), value, period

def get_time_axis(start_time, end_time, time_step, time_axis=None):
    """
        Create a list of datetimes based on an start time, end time and
        time step.  If such a list is already passed in, then this is not
        necessary.

        Often either the start_time, end_time, time_step is passed into an
        app or the time_axis is passed in directly. This function returns a
        time_axis in both situations.
    """

    #Do this import here to avoid a circular dependency
    from ..lib import units

    if time_axis is not None:
        actual_dates_axis = []
        for t in time_axis:
            #If the user has entered the time_axis with commas, remove them.
            t = t.replace(',', '').strip()
            if t == '':
                continue
            actual_dates_axis.append(get_datetime(t))
        return actual_dates_axis

    else:
        if start_time is None:
            raise HydraPluginError("A start time must be specified")
        if end_time is None:
            raise HydraPluginError("And end time must be specified")
        if time_step is None:
            raise HydraPluginError("A time-step must be specified")

        start_date = get_datetime(start_time)
        end_date = get_datetime(end_time)
        delta_t, value, output_units = parse_time_step(time_step, units_ref=units)

        time_axis = [start_date]

        value = int(value)
        while start_date < end_date:
            #Months and years are a special case, so treat them differently
            if(output_units.lower() == "mon"):
                start_date = start_date + relativedelta(months=value)
            elif (output_units.lower() == "yr"):
                start_date = start_date + relativedelta(years=value)
            else:
                start_date += timedelta(seconds=delta_t)
            time_axis.append(start_date)
        return time_axis



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
import pytest

import hydra_base
import hydra_base.exceptions

@pytest.mark.order("last")
class TestLogin:
    """ A collection of tests of the User part of the DB.
    """

    def test_login(self, client):
        """ Test adding a new user to the DB """

        user_id, session_id = client.login('root', '')

        assert user_id == 1

        assert session_id is not None
        client.logout()

    def test_logout(self, client):

        user_id, session_id = client.login('root', '')

        assert user_id == 1

        assert session_id is not None

        logout_result = client.logout()

        assert logout_result == 'OK'

    def test_get_session_user(self, client):

        #only run this for local clients, as the remote client doesn't
        # user sessions (in the test suite)
        if client.test_server is not None:
            return

        user_id, session_id = client.login('root', '')

        assert user_id == 1

        assert session_id is not None

        retrieved_user_id = client.get_session_user(session_id='i_am_not_a_session')

        assert retrieved_user_id is None

        retrieved_user_id = client.get_session_user(session_id=session_id)

        assert retrieved_user_id == user_id
        client.logout()

    def test_login_wrong_user(self, client):

        with pytest.raises(hydra_base.exceptions.HydraError):
            user_id, session_id = client.login('wrong-user!', '')

    def test_login_wrong_password(self, client):

        with pytest.raises(hydra_base.exceptions.HydraError):
            user_id, session_id = client.login('root', 'wrong-password!')


    def test_login_too_many_attempts(self, client):
        from hydra_base.lib.users import reset_failed_logins, get_failed_login_count

        for i in range(8):
            if i < 7:
                with pytest.raises(hydra_base.exceptions.HydraError):
                    user_id, session_id = client.login('root', 'wrong-password!')
            if i == 7:
                with pytest.raises(hydra_base.exceptions.HydraLoginUserMaxAttemptsExceeded):
                    user_id, session_id = client.login('root', 'wrong-password!')

        reset_failed_logins("root", flush=True)
        assert get_failed_login_count("root") == 0



from .base import *
from .scenario import *
from .attributes import *
from .dataset import *
from .notes import *
from .ownership import *
from .permissions import *
from .project import *
from .template import *
from .units import *
from .network import *
from .rule import *


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from ..base import *
from .resourceattr import ResourceAttr

__all__ = ['Resource']

class Resource:
    def get_attributes(
            self,
            include_outputs=True,
            include_inputs=True
    ):
        """
        Get the resource attributes for this resource, with 
        args:
            include_outputs (bool): Include Resource Attributes tagged with attr_is_var='Y'
            include_inputs (bool): Include Resource Attributes tagged with attr_is_var='N'
        """

        ra_qry = get_session().query(ResourceAttr).filter(
            ResourceAttr.ref_key == self.ref_key
        )

        if self.ref_key == 'NETWORK':
            ra_qry = ra_qry.filter(ResourceAttr.network_id==self.id)
        elif self.ref_key == 'NODE':
            ra_qry = ra_qry.filter(ResourceAttr.node_id==self.id)
        elif self.ref_key == 'LINK':
            ra_qry = ra_qry.filter(ResourceAttr.link_id==self.id)
        elif self.ref_key == 'GROUP':
            ra_qry = ra_qry.filter(ResourceAttr.group_id==self.id)

        if include_inputs is False or include_outputs is False:
            if include_inputs is False:
                #only include outputs
                ra_qry = ra_qry.filter(ResourceAttr.attr_is_var == 'Y')
            
            if include_outputs is False:
                #only include inputs
                ra_qry = ra_qry.filter(ResourceAttr.attr_is_var == 'N')

        return ra_qry.all()


"""dataset_id

Revision ID: dd1110144e53
Revises: 3dc489a7eac1
Create Date: 2018-04-10 22:55:42.740642

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = 'dd1110144e53'
down_revision = '3dc489a7eac1'
branch_labels = None
depends_on = None

# id         = Column(Integer(), primary_key=True, index=True, nullable=False)
# name       = Column(String(60),  nullable=False)
# type       = Column(String(60),  nullable=False)
# unit       = Column(String(60))
# hash       = Column(BIGINT(),  nullable=False, unique=True)

def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tProject

        try:
            op.alter_column('tDataset', 'dataset_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tDataset', 'data_name', new_column_name='name', existing_type=sa.String(60), nullable=False)
            op.alter_column('tDataset', 'data_type', new_column_name='type', existing_type=sa.String(60), nullable=False)
            op.alter_column('tDataset', 'data_units', new_column_name='unit', existing_type=sa.String(60))
            op.alter_column('tDataset', 'data_hash', new_column_name='hash', existing_type=sa.BIGINT(), nullable=False, unique=True)
            op.drop_column('tDataset', 'data_dimen')
            op.drop_column('tDataset', 'frequency')
            op.drop_column('tDataset', 'start_time')
        except Exception as e:
            log.exception(e)

        try:
            op.alter_column('tMetadata', 'metadata_name', new_column_name='key', existing_type=sa.String(60), nullable=False)
            op.alter_column('tMetadata', 'metadata_val', new_column_name='value', existing_type=sa.Text(1000).with_variant(mysql.TEXT(1000), 'mysql'), nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tDataset
        try:
            op.drop_table('tDataset_new')
        except:
            log.info("tDataset_new isn't there")

        try:
            # ## tDataset
            op.create_table(
                'tDataset_new',
                sa.Column('id', sa.Integer, primary_key=True),
                sa.Column('name', sa.Text(60), nullable=False),
                sa.Column('type', sa.String(60),  nullable=False),
                sa.Column('unit', sa.String(60)),
                sa.Column('hash', sa.BIGINT(),  nullable=False, unique=True),
                sa.Column('value', sa.Text(),  nullable=True),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id'),  nullable=False),
            )

            op.execute("insert into tDataset_new (id, name, type, unit, hash, value, cr_date, created_by) select dataset_id, data_name, data_type, data_units, data_hash, value, cr_date, created_by from tDataset")

            op.rename_table('tDataset','tDataset_old')
            op.rename_table('tDataset_new', 'tDataset')
            op.drop_table('tDataset_old')

        except Exception as e:
            log.exception(e)

        # ## tMetadata
        try:
            op.drop_table('tMetadata_new')
        except:
            log.info("tMetadata_new isn't there")

        try:
            # ## tMetadata
            op.create_table(
                'tMetadata_new',
                sa.Column('dataset_id', sa.Integer, primary_key=True, index=True),
                sa.Column('key', sa.String(60), nullable=False, primary_key=True),
                sa.Column('value', sa.Text(),  nullable=False),
            )

            op.execute("insert into tMetadata_new (dataset_id, key, value) select dataset_id, metadata_name, metadata_val from tMetadata")

            op.rename_table('tMetadata','tMetadata_old')
            op.rename_table('tMetadata_new', 'tMetadata')
            op.drop_table('tMetadata_old')

        except Exception as e:
            log.exception(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tProject

        try:
            op.alter_column('tDataset', 'id', new_column_name='dataset_id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tDataset', 'name', new_column_name='dasta_name', existing_type=sa.String(60), nullable=False)
            op.alter_column('tDataset', 'type', new_column_name='data_type', existing_type=sa.String(60), nullable=False)
            op.alter_column('tDataset', 'unit', new_column_name='data_units', existing_type=sa.String(60))
            op.add_column('tDataset', 'data_dimen', sa.String(60))
        except Exception as e:
            log.exception(e)

        try:
            op.alter_column('tMetadata', 'key', new_column_name='metadata_name', existing_type=sa.String(60), nullable=False)
            op.alter_column('tMetadata', 'value', new_column_name='metadata_val', existing_type=sa.Text().with_variant(mysql.TEXT(4294967295), 'mysql'), nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tDataset
        try:
            op.drop_table('tDataset_new')
        except:
            log.info("tDataset_new isn't there")

        try:
            # ## tDataset
            op.create_table(
                'tDataset_new',
                sa.Column('dataset_id', sa.Integer, primary_key=True),
                sa.Column('data_name', sa.Text(60), nullable=False),
                sa.Column('data_type', sa.String(60),  nullable=False),
                sa.Column('data_units', sa.String(60)),
                sa.Column('data_dimen', sa.String(60)),
                sa.Column('data_hash', sa.BIGINT(),  nullable=False, unique=True),
                sa.Column('value', sa.Text(),  nullable=True),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id'),  nullable=False),
                sa.Column('start_time', sa.Text(60),  nullable=True),
                sa.Column('frequency', sa.Text(10),  nullable=True),
            )

            op.execute("insert into tDataset_new (dataset_id, data_name, data_type, data_units, data_hash, value, cr_date, created_by) select id, name, type, unit, hash, value, cr_date, created_by from tDataset")

            op.rename_table('tDataset','tDataset_old')
            op.rename_table('tDataset_new', 'tDataset')
            op.drop_table('tDataset_old')

        except Exception as e:
            log.exception(e)
        # ## tMetadata
        try:
            op.drop_table('tMetadata_new')
        except:
            log.info("tMetadata_new isn't there")
        try:
            # ## tMetadata
            op.create_table(
                'tMetadata_new',
                sa.Column('dataset_id', sa.Integer, primary_key=True, index=True),
                sa.Column('metadata_name', sa.String(60), nullable=False, primary_key=True),
                sa.Column('metadata_val', sa.Text(),  nullable=False),
            )

            op.execute("insert into tMetadata_new (dataset_id, metadata_name, metadata_val) select dataset_id, key, value from tMetadata")

            op.rename_table('tMetadata','tMetadata_old')
            op.rename_table('tMetadata_new', 'tMetadata')
            op.drop_table('tMetadata_old')

        except Exception as e:
            log.exception(e)
